import math
from abc import abstractmethod
from typing import Any, Generic, TypeVar

import torch
from accelerate import Accelerator
from transformers import GenerationConfig, PreTrainedModel, PreTrainedTokenizerBase

from turbo_alignment.dataset.base import BaseDataset
from turbo_alignment.dataset.base.models import DatasetRecord
from turbo_alignment.settings.generators.chat import CustomChatGenerationSettings
from turbo_alignment.settings.generators.outputs.base import BaseInferenceOutput
from turbo_alignment.settings.tf.generation import GeneratorTransformersSettings

InferenceOutputT = TypeVar('InferenceOutputT', bound=BaseInferenceOutput)
DatasetRecordT = TypeVar('DatasetRecordT', bound=DatasetRecord)


class BaseGenerator(Generic[DatasetRecordT, InferenceOutputT]):
    def __init__(
        self,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        accelerator: Accelerator | None = None,
        batch: int = 1,
    ) -> None:
        if accelerator is not None:
            self._model = accelerator.prepare_model(model, device_placement=True, evaluation_mode=True)
        else:
            self._model = model

        self._tokenizer = tokenizer
        self._accelerator = accelerator
        self._batch = batch

    @abstractmethod
    def _generate_from_batch(
        self,
        records: list[dict[str, Any]],
        original_records: list[DatasetRecordT],
        dataset_name: str,
    ) -> list[InferenceOutputT]:
        ...

    @property
    def device(self):
        return self._accelerator.device if self._accelerator is not None else self._model.device

    def generate_from_dataset(self, dataset: BaseDataset) -> list[InferenceOutputT]:
        input_records = [dataset[idx] for idx in range(len(dataset))]

        records_batches = [
            input_records[i * self._batch : (i + 1) * self._batch]
            for i in range(math.ceil(len(dataset) / self._batch))
        ]

        original_records_batches = [
            [dataset.get_original_record_by_id(r['id']) for r in batch] for batch in records_batches
        ]

        if self._accelerator is None:
            generations = []
            for i, (records_batch, original_records_batch) in enumerate(
                zip(records_batches, original_records_batches)
            ):
                generations.append(
                    self._generate_from_batch(
                        records_batch,
                        original_records_batch,
                        dataset.source.name,
                    )
                )
        else:
            with self._accelerator.split_between_processes(
                list(zip(records_batches, original_records_batches)), apply_padding=True
            ) as accelerator_records:
                generations = [
                    self._generate_from_batch(
                        records_batch,
                        original_records_batch,
                        dataset.source.name,
                    )
                    for records_batch, original_records_batch in accelerator_records
                ][: len(records_batches)]

        return sum(generations, [])


class ChatGeneratorBase(BaseGenerator, Generic[DatasetRecordT, InferenceOutputT]):
    def __init__(
        self,
        transformers_settings: GeneratorTransformersSettings,
        custom_generation_settings: CustomChatGenerationSettings,
        tokenizer: PreTrainedTokenizerBase,
        return_logits: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(tokenizer=tokenizer, **kwargs)

        self._return_logits = return_logits

        self._transformers_generator_parameters = GenerationConfig(
            bos_token_id=self._tokenizer.bos_token_id,
            **transformers_settings.dict(),
        )

        self._custom_generation_settings = custom_generation_settings

    @abstractmethod
    def _generate_from_single_record(
        self,
        record: dict[str, torch.Tensor],
        original_record: DatasetRecordT,
        dataset_name: str,
    ) -> InferenceOutputT:
        ...

    @abstractmethod
    def _generate_from_batch_records(
        self,
        records: list[dict[str, torch.Tensor]],
        original_records: list[DatasetRecordT],
        dataset_name: str,
    ) -> list[InferenceOutputT]:
        ...

    def _generate_from_batch(
        self, records: list[dict[str, Any]], original_records: list[DatasetRecordT], dataset_name: str
    ) -> list[InferenceOutputT]:
        if self._custom_generation_settings.batch > 1:
            if self._transformers_generator_parameters.num_beams != 1:
                raise ValueError('You can not use batch generation with num_beams != 1')

            self._tokenizer.padding_side = 'left'
            self._tokenizer.pad_token_id = self._tokenizer.pad_token_id

            return self._generate_from_batch_records(records, original_records, dataset_name)

        return [
            self._generate_from_single_record(record, original_record, dataset_name)
            for record, original_record in zip(records, original_records)
        ]

    @staticmethod
    def _postprocess(input_indices: torch.Tensor, output_indices: torch.Tensor, remove_prompt: bool) -> torch.Tensor:
        if remove_prompt:
            return output_indices[:, input_indices.shape[1] :].cpu()
        return output_indices.cpu()

    def _decode(self, token_indices: torch.Tensor) -> list[str]:
        return self._tokenizer.batch_decode(
            token_indices, skip_special_tokens=self._custom_generation_settings.skip_special_tokens
        )



from abc import abstractmethod
from typing import Generic, Iterable, TypeVar

from transformers import (
    PreTrainedModel,
    PreTrainedTokenizerBase,
    TrainerCallback,
    TrainerControl,
    TrainerState,
    TrainingArguments,
)

from turbo_alignment.dataset.base import BaseDataset
from turbo_alignment.metrics.metric import Metric
from turbo_alignment.settings.cherry_pick import CherryPickSettings
from turbo_alignment.settings.metric import MetricResults

InferenceDatasetT = TypeVar('InferenceDatasetT', bound=BaseDataset)


class CherryPickCallbackBase(TrainerCallback, Generic[InferenceDatasetT]):
    def __init__(
        self,
        cherry_pick_settings: CherryPickSettings,
        datasets: Iterable[InferenceDatasetT],
        metrics: list[Metric],
    ):
        super().__init__()
        self._cherry_pick_settings = cherry_pick_settings
        self._metrics = metrics
        self._datasets = datasets

    @abstractmethod
    def _get_dataset_metrics(
        self,
        dataset: InferenceDatasetT,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> list[MetricResults]:
        ...

    def on_evaluate(
        self,
        _args: TrainingArguments,
        state: TrainerState,
        _control: TrainerControl,
        **kwargs,
    ) -> list[list[MetricResults]]:
        tokenizer: PreTrainedTokenizerBase = kwargs.pop('tokenizer', None)
        if tokenizer is None:
            raise ValueError('Tokenizer is None')

        model: PreTrainedModel = kwargs.pop('model', None)
        if model is None:
            raise ValueError('Model is None')

        model.eval()

        dataset_metrics = []

        for dataset in self._datasets:
            dataset_metrics.append(
                self._get_dataset_metrics(
                    dataset=dataset,
                    model=model,
                    tokenizer=tokenizer,
                    **kwargs,
                ),
            )

        model.train()

        return dataset_metrics



# mypy: ignore-errors
# pylint: skip-file
import collections.abc
import inspect
import logging
from copy import deepcopy
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from turbo_alignment.common.registry.checks import ConfigurationError
from turbo_alignment.common.registry.lazy import Lazy
from turbo_alignment.common.registry.params import Params

logger = logging.getLogger(__name__)

T = TypeVar('T', bound='FromParams')

_NO_DEFAULT = inspect.Parameter.empty


def takes_arg(obj, arg: str) -> bool:
    if inspect.isclass(obj):
        signature = inspect.signature(obj.__init__)
    elif inspect.ismethod(obj) or inspect.isfunction(obj):
        signature = inspect.signature(obj)
    else:
        raise ConfigurationError(f'object {obj} is not callable')
    return arg in signature.parameters


def takes_kwargs(obj) -> bool:
    if inspect.isclass(obj):
        signature = inspect.signature(obj.__init__)
    elif inspect.ismethod(obj) or inspect.isfunction(obj):
        signature = inspect.signature(obj)
    else:
        raise ConfigurationError(f'object {obj} is not callable')
    return any(p.kind == inspect.Parameter.VAR_KEYWORD for p in signature.parameters.values())  # type: ignore


def can_construct_from_params(type_: Type) -> bool:
    if type_ in [str, int, float, bool]:
        return True
    origin = getattr(type_, '__origin__', None)
    if origin == Lazy:
        return True
    if origin:
        if hasattr(type_, 'from_params'):
            return True
        args = getattr(type_, '__args__')
        return all(can_construct_from_params(arg) for arg in args)

    return hasattr(type_, 'from_params')


def is_base_registrable(cls) -> bool:
    from turbo_alignment.common.registry.registrable import (
        Registrable,  # import here to avoid circular imports
    )

    if not issubclass(cls, Registrable):
        return False
    method_resolution_order = inspect.getmro(cls)[1:]
    for base_class in method_resolution_order:
        if issubclass(base_class, Registrable) and base_class is not Registrable:
            return False
    return True


def remove_optional(annotation: type):
    origin = getattr(annotation, '__origin__', None)
    args = getattr(annotation, '__args__', ())

    if origin == Union:
        return Union[tuple([arg for arg in args if arg != type(None)])]  # noqa: E721
    return annotation


def infer_constructor_params(
    cls: Type[T], constructor: Union[Callable[..., T], Callable[[T], None]] = None
) -> Dict[str, inspect.Parameter]:
    if constructor is None:
        constructor = cls.__init__
    return infer_method_params(cls, constructor)


infer_params = infer_constructor_params  # Legacy name


def infer_method_params(cls: Type[T], method: Callable) -> Dict[str, inspect.Parameter]:
    signature = inspect.signature(method)
    parameters = dict(signature.parameters)

    has_kwargs = False
    var_positional_key = None
    for param in parameters.values():
        if param.kind == param.VAR_KEYWORD:
            has_kwargs = True
        elif param.kind == param.VAR_POSITIONAL:
            var_positional_key = param.name

    if var_positional_key:
        del parameters[var_positional_key]

    if not has_kwargs:
        return parameters

    super_class = None
    for super_class_candidate in cls.mro()[1:]:
        if issubclass(super_class_candidate, FromParams):
            super_class = super_class_candidate
            break
    if super_class:
        super_parameters = infer_params(super_class)
    else:
        super_parameters = {}

    return {**super_parameters, **parameters}


def create_kwargs(constructor: Callable[..., T], cls: Type[T], params: Params, **extras) -> Dict[str, Any]:
    kwargs: Dict[str, Any] = {}

    parameters = infer_params(cls, constructor)
    accepts_kwargs = False

    for param_name, param in parameters.items():
        if param_name == 'self':
            continue

        if param.kind == param.VAR_KEYWORD:
            accepts_kwargs = True
            continue

        annotation = remove_optional(param.annotation)

        explicitly_set = param_name in params
        # FIXME
        constructed_arg = pop_and_construct_arg(cls.__name__, param_name, annotation, param.default, params, **extras)

        if explicitly_set or constructed_arg is not param.default:
            kwargs[param_name] = constructed_arg

    if accepts_kwargs:
        kwargs.update(params)
    else:
        params.assert_empty(cls.__name__)
    return kwargs


def create_extras(cls: Type[T], extras: Dict[str, Any]) -> Dict[str, Any]:
    subextras: Dict[str, Any] = {}
    if hasattr(cls, 'from_params'):
        from_params_method = cls.from_params
    else:
        from_params_method = cls
    if takes_kwargs(from_params_method):
        subextras = extras
    else:
        subextras = {k: v for k, v in extras.items() if takes_arg(from_params_method, k)}
    return subextras


def construct_arg(
    class_name: str,
    argument_name: str,
    popped_params: Params,
    annotation: Type,
    default: Any,
    **extras,
) -> Any:
    origin = getattr(annotation, '__origin__', None)
    args = getattr(annotation, '__args__', [])

    optional = default != _NO_DEFAULT

    if hasattr(annotation, 'from_params'):
        if popped_params is default:
            return default
        if popped_params is not None:
            subextras = create_extras(annotation, extras)

            if isinstance(popped_params, str):
                popped_params = Params({'type': popped_params})
            elif isinstance(popped_params, dict):
                popped_params = Params(popped_params)
            result = annotation.from_params(params=popped_params, **subextras)

            return result
        if not optional:
            raise ConfigurationError(f'expected key {argument_name} for {class_name}')
        return default

    if annotation in {int, bool}:
        if type(popped_params) in {int, bool}:
            return annotation(popped_params)
        raise TypeError(f'Expected {argument_name} to be a {annotation.__name__}.')
    if annotation == str:
        if isinstance(popped_params, str) or isinstance(popped_params, Path):
            return str(popped_params)  # type: ignore
        raise TypeError(f'Expected {argument_name} to be a string.')
    if annotation == float:
        if type(popped_params) in {int, float}:
            return popped_params
        raise TypeError(f'Expected {argument_name} to be numeric.')

    if (
        origin in {collections.abc.Mapping, Mapping, Dict, dict}
        and len(args) == 2
        and can_construct_from_params(args[-1])
    ):
        value_cls = annotation.__args__[-1]
        value_dict = {}
        if not isinstance(popped_params, Mapping):
            raise TypeError(f'Expected {argument_name} to be a Mapping (probably a dict or a Params object).')

        for key, value_params in popped_params.items():
            value_dict[key] = construct_arg(
                str(value_cls),
                argument_name + '.' + key,
                value_params,
                value_cls,
                _NO_DEFAULT,
                **extras,
            )

        return value_dict

    if origin in (Tuple, tuple) and all(can_construct_from_params(arg) for arg in args):
        value_list = []

        for i, (value_cls, value_params) in enumerate(zip(annotation.__args__, popped_params)):
            value = construct_arg(
                str(value_cls),
                argument_name + f'.{i}',
                value_params,
                value_cls,
                _NO_DEFAULT,
                **extras,
            )
            value_list.append(value)

        return tuple(value_list)

    if origin in (Set, set) and len(args) == 1 and can_construct_from_params(args[0]):
        value_cls = annotation.__args__[0]

        value_set = set()

        for i, value_params in enumerate(popped_params):
            value = construct_arg(
                str(value_cls),
                argument_name + f'.{i}',
                value_params,
                value_cls,
                _NO_DEFAULT,
                **extras,
            )
            value_set.add(value)

        return value_set

    if origin == Union:
        backup_params = deepcopy(popped_params)

        error_chain: Optional[Exception] = None
        for arg_annotation in args:
            try:
                return construct_arg(
                    str(arg_annotation),
                    argument_name,
                    popped_params,
                    arg_annotation,
                    default,
                    **extras,
                )
            except (ValueError, TypeError, ConfigurationError, AttributeError) as e:
                popped_params = deepcopy(backup_params)
                e.args = (f'While constructing an argument of type {arg_annotation}',) + e.args
                e.__cause__ = error_chain
                error_chain = e

        config_error = ConfigurationError(f'Failed to construct argument {argument_name} with type {annotation}.')
        config_error.__cause__ = error_chain
        raise config_error
    if origin == Lazy:
        if popped_params is default:
            return default

        value_cls = args[0]
        subextras = create_extras(value_cls, extras)
        return Lazy(value_cls, params=deepcopy(popped_params), constructor_extras=subextras)  # type: ignore

    if (
        origin in {collections.abc.Iterable, Iterable, List, list}
        and len(args) == 1
        and can_construct_from_params(args[0])
    ):
        value_cls = annotation.__args__[0]

        value_list = []

        for i, value_params in enumerate(popped_params):
            value = construct_arg(
                str(value_cls),
                argument_name + f'.{i}',
                value_params,
                value_cls,
                _NO_DEFAULT,
                **extras,
            )
            value_list.append(value)

        return value_list

    if isinstance(popped_params, Params):
        return popped_params.as_dict()
    return popped_params


class FromParams:
    @classmethod
    def from_params(
        cls: Type[T],
        params: Params,
        constructor_to_call: Callable[..., T] = None,
        constructor_to_inspect: Union[Callable[..., T], Callable[[T], None]] = None,
        **extras,
    ) -> T:
        from turbo_alignment.common.registry.registrable import (
            Registrable,  # import here to avoid circular imports
        )

        logger.debug(
            f"instantiating class {cls} from params {getattr(params, 'params', params)} "
            f'and extras {set(extras.keys())}'
        )

        if params is None:
            return None

        if isinstance(params, str):
            params = Params({'type': params})

        if not isinstance(params, Params):
            raise ConfigurationError(
                'from_params was passed a `params` object that was not a `Params`. This probably '
                'indicates malformed parameters in a configuration file, where something that '
                'should have been a dictionary was actually a list, or something else. '
                f'This happened when constructing an object of type {cls}.'
            )

        registered_subclasses = Registrable._registry.get(cls)

        if is_base_registrable(cls) and registered_subclasses is None:
            raise ConfigurationError(
                'Tried to construct an abstract Registrable base class that has no registered '
                'concrete types. This might mean that you need to use --include-package to get '
                'your concrete classes actually registered.'
            )

        if registered_subclasses is not None and not constructor_to_call:
            # FIXME
            as_registrable = cast(Type[Registrable], cls)
            default_to_first_choice = as_registrable.default_implementation is not None
            # FIXME
            choice = params.pop_choice(
                'type',
                choices=as_registrable.list_available(),
                default_to_first_choice=default_to_first_choice,
            )
            subclass, constructor_name = as_registrable.resolve_class_name(choice)
            if not constructor_name:
                constructor_to_inspect = subclass.__init__
                constructor_to_call = subclass  # type: ignore
            else:
                constructor_to_inspect = cast(Callable[..., T], getattr(subclass, constructor_name))
                constructor_to_call = constructor_to_inspect

            if hasattr(subclass, 'from_params'):
                extras = create_extras(subclass, extras)
                retyped_subclass = cast(Type[T], subclass)
                return retyped_subclass.from_params(
                    params=params,
                    constructor_to_call=constructor_to_call,
                    constructor_to_inspect=constructor_to_inspect,
                    **extras,
                )
            return subclass(**params)  # type: ignore
        if not constructor_to_inspect:
            constructor_to_inspect = cls.__init__
        if not constructor_to_call:
            constructor_to_call = cls

        if constructor_to_inspect == object.__init__:
            kwargs: Dict[str, Any] = {}
            params.assert_empty(cls.__name__)
        else:
            constructor_to_inspect = cast(Callable[..., T], constructor_to_inspect)
            kwargs = create_kwargs(constructor_to_inspect, cls, params, **extras)

        return constructor_to_call(**kwargs)  # type: ignore

    def to_params(self) -> Params:
        """
        Returns a `Params` object that can be used with `.from_params()` to recreate an
        object just like it.

        This relies on `_to_params()`. If you need this in your custom `FromParams` class,
        override `_to_params()`, not this method.
        """

        def replace_object_with_params(o: Any) -> Any:
            if isinstance(o, FromParams):
                return o.to_params()
            if isinstance(o, List):
                return [replace_object_with_params(i) for i in o]
            if isinstance(o, Set):
                return {replace_object_with_params(i) for i in o}
            if isinstance(o, Dict):
                return {key: replace_object_with_params(value) for key, value in o.items()}
            return o

        return Params(replace_object_with_params(self._to_params()))

    def _to_params(self) -> Dict[str, Any]:
        """
        Returns a dictionary of parameters that, when turned into a `Params` object and
        then fed to `.from_params()`, will recreate this object.

        You don't need to implement this all the time. AllenNLP will let you know if you
        need it.
        """
        raise NotImplementedError()



import torch
from torch import nn
from torch.nn import Module
from transformers import (
    AutoModel,
    AutoModelForCausalLM,
    GenerationConfig,
    PreTrainedModel,
)

from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.modeling.rag.rag_output import (
    RetrieveAugLMMarginOutput,
    RetrieveAugLMOutput,
)
from turbo_alignment.modeling.rag.rag_tokenizer import RagTokenizer
from turbo_alignment.modeling.rag.retriever_model import RagRetriever
from turbo_alignment.modeling.rag.utils import (
    get_question_embeddings,
    join_query_and_docs,
)
from turbo_alignment.settings.rag_model import RAGPreTrainedModelSettings

logger = get_project_logger()


class RagModel(Module):
    def __init__(
        self,
        generator: AutoModelForCausalLM,
        question_encoder: AutoModel,
        config: RAGPreTrainedModelSettings,
        tokenizer: RagTokenizer,
    ):
        super().__init__()
        self.rag_config = config

        self.generator = generator
        self.question_encoder = question_encoder
        self.retriever = RagRetriever(config=config, rag_tokenizer=tokenizer)

        self.ctx_encoder = None
        self.context_encoder_training = False
        self.tokenizer = tokenizer

    def forward(
        self,
        input_ids: torch.LongTensor,
        attention_mask: torch.Tensor,
        labels: torch.Tensor,
    ) -> RetrieveAugLMOutput:
        input_ids = input_ids.to(self.generator.device)
        labels = labels.to(self.generator.device)

        # decode query with generator tokenizer
        decoded_inputs = self.tokenizer.batch_decode(input_ids, skip_special_tokens=True)

        # TODO HACK
        decoded_inputs = [i.split('<RS><bot>')[0] for i in decoded_inputs]

        decoded_inputs = [f'{self.rag_config.retrieval_settings.prefix} {inp}' for inp in decoded_inputs]
        # encode query with retriever tokenizer
        retriever_tokenizer_output = self.tokenizer.question_encoder(
            decoded_inputs,
            max_length=self.rag_config.retrieval_settings.query_encoder_max_length,
            padding=True,
            truncation=True,
            return_tensors='pt',
        )
        retriever_input_ids = retriever_tokenizer_output['input_ids'].to(self.question_encoder.device)
        retriever_attn_mask = retriever_tokenizer_output['attention_mask'].to(self.question_encoder.device)

        # get query embeddings
        question_enc_outputs = self.question_encoder(
            retriever_input_ids,
            attention_mask=retriever_attn_mask,
            return_dict=True,
            output_hidden_states=True,
        )
        question_encoder_last_hidden_state = get_question_embeddings(question_enc_outputs, retriever_attn_mask)

        # retrieve ONE document, it will already be tokenized with generator tokenizer
        retriever_outputs = self.retriever(
            retriever_input_ids,
            question_encoder_last_hidden_state.cpu().detach().to(torch.float32).numpy(),
            n_docs=self.rag_config.retrieval_settings.n_docs,
            return_tensors='pt',
        )
        doc_input_ids, doc_attention_mask, retrieved_doc_embeds, retrieved_doc_ids = (
            retriever_outputs['context_input_ids'],
            retriever_outputs['context_attention_mask'],
            retriever_outputs['retrieved_doc_embeds'],
            retriever_outputs['doc_ids'],
        )
        doc_input_ids = doc_input_ids.to(input_ids)
        doc_attention_mask = doc_attention_mask.to(input_ids)

        # join document and query input ids
        joined_input_ids, joined_attention_mask, joined_labels = join_query_and_docs(
            input_ids, doc_input_ids, self.tokenizer.pad_token_id, attention_mask, doc_attention_mask, labels
        )

        # set to correct device
        retrieved_doc_embeds = retrieved_doc_embeds.to(question_encoder_last_hidden_state)
        joined_input_ids = joined_input_ids.to(input_ids)
        joined_attention_mask = joined_attention_mask.to(input_ids)

        # compute doc_scores
        doc_scores = torch.bmm(
            question_encoder_last_hidden_state.unsqueeze(1), retrieved_doc_embeds.transpose(1, 2)
        ).squeeze(1)

        gen_outputs = self.generator(
            input_ids=joined_input_ids,
            attention_mask=joined_attention_mask,
            return_dict=True,
        )

        question_enc_hidden_states = question_enc_outputs.hidden_states
        question_enc_attentions = question_enc_outputs.attentions

        return RetrieveAugLMOutput(
            logits=gen_outputs.logits,
            labels=joined_labels,
            doc_scores=doc_scores,
            past_key_values=gen_outputs.past_key_values,
            joined_input_ids=joined_input_ids,
            joined_attention_mask=joined_attention_mask,
            retrieved_doc_embeds=retrieved_doc_embeds,
            retrieved_doc_ids=retrieved_doc_ids,
            question_encoder_last_hidden_state=question_encoder_last_hidden_state,
            question_enc_hidden_states=question_enc_hidden_states,
            question_enc_attentions=question_enc_attentions,
            generator_dec_hidden_states=gen_outputs.hidden_states,
            generator_dec_attentions=gen_outputs.attentions,
        )


class RagSequenceForGeneration(Module):
    def __init__(
        self,
        model_settings: RAGPreTrainedModelSettings,
        generator: AutoModelForCausalLM,
        question_encoder: AutoModel,
        tokenizer: RagTokenizer,
    ):
        '''
        Implementation of RAG-Sequence from the paper:
        Retrieval-Augmented Generation for Knowledge-Intensive NLP Tasks
        https://arxiv.org/pdf/2005.11401
        '''
        super().__init__()
        self.rag_config = model_settings
        self.generation_config = GenerationConfig
        self.tokenizer = tokenizer
        self.rag = RagModel(
            config=self.rag_config, generator=generator, question_encoder=question_encoder, tokenizer=tokenizer
        )

        # this config is needed for Trainer to work, rag_config is not PretrainedConfig
        self.config = self.rag.generator.config

        self.device = self.rag.generator.device

    def set_retriever(self, retriever: RagRetriever) -> None:
        self.rag.retriever = retriever

    def set_context_encoder_for_training(self, ctx_encoder: PreTrainedModel) -> None:
        self.rag.context_encoder_training = True
        self.rag.ctx_encoder = ctx_encoder

    def resize_token_embeddings(self, emb_size: int) -> None:
        self.rag.generator.resize_token_embeddings(emb_size)
        self.rag.question_encoder.resize_token_embeddings(len(self.tokenizer.question_encoder))

    def forward(
        self,
        input_ids: torch.LongTensor,
        attention_mask: torch.Tensor | None = None,
        labels: torch.LongTensor | None = None,
        **_kwargs,
    ) -> RetrieveAugLMMarginOutput:
        outputs = self.rag(
            input_ids=input_ids,
            attention_mask=attention_mask,
            labels=labels,
        )

        logits = outputs.logits
        labels = outputs.labels
        joined_input_ids = outputs.joined_input_ids

        loss = self.get_nll(
            logits,
            outputs.doc_scores,
            target=joined_input_ids,
            labels=labels,
            n_docs=self.rag_config.retrieval_settings.n_docs,
        )

        out = RetrieveAugLMMarginOutput(
            loss=loss,
            logits=logits,
            doc_scores=outputs.doc_scores,
            past_key_values=outputs.past_key_values,
            context_input_ids=outputs.joined_input_ids,
            context_attention_mask=outputs.joined_attention_mask,
            retrieved_doc_embeds=outputs.retrieved_doc_embeds,
            retrieved_doc_ids=outputs.retrieved_doc_ids,
            question_encoder_last_hidden_state=outputs.question_encoder_last_hidden_state,
            question_enc_hidden_states=outputs.question_enc_hidden_states,  # nones from here
            question_enc_attentions=outputs.question_enc_attentions,
            generator_enc_last_hidden_state=outputs.generator_enc_last_hidden_state,
            generator_enc_hidden_states=outputs.generator_enc_hidden_states,
            generator_enc_attentions=outputs.generator_enc_attentions,
            generator_dec_hidden_states=outputs.generator_dec_hidden_states,
            generator_dec_attentions=outputs.generator_dec_attentions,
            generator_cross_attentions=outputs.generator_cross_attentions,
        )
        return out

    def get_nll(
        self,
        seq_logits: torch.Tensor,
        doc_scores: torch.Tensor,
        target: torch.Tensor,
        labels: torch.Tensor,
        n_docs: int,
    ) -> torch.Tensor:
        '''
        seq_logits: [batch_size*n_docs, seq_len, vocab_size]
        doc_score: [batch_size, n_docs]
        targets: [batch_size*n_docs, seq_len]
        labels: [batch_size*n_docs, seq_len]
        '''

        # https://github.com/huggingface/transformers/blob/main/src/transformers/models/rag/modeling_rag.py#L1057

        batch_size = seq_logits.shape[0] // n_docs

        def _mask_pads(ll):
            ignore_index = -100
            pad_mask = labels.eq(ignore_index)
            if pad_mask.any():
                ll.masked_fill_(pad_mask, 0.0)
            return ll.squeeze(-1)

        # seq_logits dim = (batch*n_docs, seq_len , vocab_size)
        seq_logprobs = nn.functional.log_softmax(seq_logits, dim=-1).view(
            batch_size, n_docs, -1, seq_logits.size(-1)
        )  # batch_size , n_docs , seq_len , vocab_size

        doc_logprobs = nn.functional.log_softmax(doc_scores, dim=-1)

        # calculate loss
        seq_logprobs = seq_logprobs.flatten(0, 1)  # batch_size*n_docs , seq_len , vocab_size
        target = target.unsqueeze(-1)
        assert target.dim() == seq_logprobs.dim()

        seq_logprobs = seq_logprobs.gather(dim=-1, index=target)

        seq_logprobs = seq_logprobs.squeeze(-1)

        seq_logprobs_masked = _mask_pads(seq_logprobs)

        # sum over tokens, exclude bos while scoring
        seq_logprobs_masked = seq_logprobs_masked.reshape((batch_size, n_docs, -1))  # batch_size, n_docs, seq_len
        seq_ll = seq_logprobs_masked.sum(2)  # batch_size, n_docs

        # marginalize over docs
        ll = seq_ll + doc_logprobs

        ll = ll.logsumexp(1)  # logsumexp over docs

        loss = -ll.mean()

        return loss

    @property
    def retriever(self) -> RagRetriever:
        return self.rag.retriever

    @property
    def generator(self):
        return self.rag.generator

    @property
    def question_encoder(self):
        return self.rag.question_encoder

    def prepare_inputs_for_generation(self, *args, **kwargs):
        return self.generator.prepare_inputs_for_generation(*args, **kwargs)

    @torch.no_grad()
    def generate(
        self,
        inputs: torch.LongTensor,
        generation_config: GenerationConfig | None = None,
        **kwargs,
    ) -> tuple[torch.LongTensor, torch.LongTensor, torch.Tensor]:
        # TODO remove code duplicate with forward

        inputs = inputs.to(self.question_encoder.device)

        # decode query with generator tokenizer
        decoded_inputs = self.tokenizer.batch_decode(inputs)
        # encode query with retriever tokenizer
        retriever_tokenizer_output = self.tokenizer.question_encoder(
            decoded_inputs,
            max_length=self.rag_config.retrieval_settings.query_encoder_max_length,
            padding=True,
            truncation=True,
            return_tensors='pt',
        )
        retriever_input_ids = retriever_tokenizer_output['input_ids'].to(self.question_encoder.device)
        retriever_attn_mask = retriever_tokenizer_output['attention_mask'].to(self.question_encoder.device)

        question_encoder_output = self.question_encoder(
            retriever_input_ids, attention_mask=retriever_attn_mask, output_hidden_states=True
        )
        question_hidden_states = get_question_embeddings(question_encoder_output, retriever_attn_mask)

        retriever_output = self.retriever(
            retriever_input_ids,
            question_hidden_states.cpu().detach().to(torch.float32).numpy(),
            n_docs=self.rag_config.retrieval_settings.n_docs,
            return_tensors='pt',
        )

        # this in unnecessary, computing doc scores only for wandb table
        retrieved_doc_embeds = retriever_output['retrieved_doc_embeds'].to(question_hidden_states)
        doc_scores = torch.bmm(question_hidden_states.unsqueeze(1), retrieved_doc_embeds.transpose(1, 2)).squeeze(1)

        doc_input_ids = retriever_output['context_input_ids'].to(self.question_encoder.device)
        joined_input_ids, _, _ = join_query_and_docs(
            input_ids=inputs, doc_input_ids=doc_input_ids, pad_token_id=self.tokenizer.pad_token_id
        )

        batch_size = retriever_input_ids.shape[0]
        assert batch_size == 1, 'Can not generate using batches for now'

        output_sequences = self.generator.generate(
            input_ids=joined_input_ids,
            generation_config=generation_config,
            pad_token_id=self.tokenizer.pad_token_id,
            tokenizer=kwargs.get('tokenizer', None),
        )
        # TODO chose max-prob sequence with accounting for doc probs
        only_answer_output = output_sequences[:, joined_input_ids.shape[-1] :]
        return only_answer_output, doc_input_ids, doc_scores



from turbo_alignment.common.registry import Registrable
from turbo_alignment.settings.datasets import DatasetType


class DatasetRegistry(Registrable):
    ...


@DatasetRegistry.register(DatasetType.DDPO)
class DDPODatasetTypeRegistry(Registrable):
    ...


@DatasetRegistry.register(DatasetType.PAIR_PREFERENCES)
class PairPreferenceDatasetTypeRegistry(Registrable):
    ...


@DatasetRegistry.register(DatasetType.KTO)
class KTODatasetTypeRegistry(Registrable):
    ...


@DatasetRegistry.register(DatasetType.CHAT)
class ChatDatasetTypeRegistry(Registrable):
    ...


@DatasetRegistry.register(DatasetType.CLASSIFICATION)
class ClassificationDatasetTypeRegistry(Registrable):
    ...


@DatasetRegistry.register(DatasetType.SAMPLING)
class SamplingRMDatasetTypeRegistry(Registrable):
    ...


@DatasetRegistry.register(DatasetType.MULTIMODAL)
class MultimodalDatasetTypeRegistry(Registrable):
    ...



import torch
from transformers import PreTrainedTokenizerBase

from turbo_alignment.metrics.metric import Metric
from turbo_alignment.settings.metric import ElementWiseScores, MetricResults, MetricType


@Metric.register(MetricType.LENGTH)
class LengthMetric(Metric):
    def compute(self, **kwargs) -> list[MetricResults]:
        tokenizer: PreTrainedTokenizerBase = kwargs.get('tokenizer', None)
        answer_tokens_ids: list[torch.Tensor] = kwargs.get('answer_tokens_ids', None)
        dataset_name: str = kwargs.get('dataset_name', '')

        if answer_tokens_ids is None:
            raise ValueError('answer_tokens_ids should not be None')

        if tokenizer is None:
            raise ValueError('tokenizer should not be None')

        return [
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(
                        label=dataset_name + '@@' + 'length',
                        values=[
                            (answer_tokens != tokenizer.pad_token_id).sum().item()
                            for answer_tokens in answer_tokens_ids
                        ],
                    )
                ],
                need_average=need_average,
            )
            for need_average in self._settings.need_average
        ]



# mypy: ignore-errors
# pylint: skip-file

import functools
import importlib
from typing import Callable

import torch
import triton
from packaging.version import Version


def ensure_contiguous(fn):
    @functools.wraps(fn)
    def wrapper(ctx, *args, **kwargs):
        def maybe_to_contiguous(x):
            return x.contiguous() if isinstance(x, torch.Tensor) else x

        args = [maybe_to_contiguous(arg) for arg in args]
        kwargs = {k: maybe_to_contiguous(v) for k, v in kwargs.items()}
        return fn(ctx, *args, **kwargs)

    return wrapper


def calculate_settings(n):
    MAX_FUSED_SIZE = 65536
    BLOCK_SIZE = triton.next_power_of_2(n)
    if BLOCK_SIZE > MAX_FUSED_SIZE:
        raise RuntimeError(
            f'Cannot launch Triton kernel since n = {n} exceeds '
            f'the recommended Triton blocksize = {MAX_FUSED_SIZE}.'
        )

    num_warps = 4
    if BLOCK_SIZE >= 32768:
        num_warps = 32
    elif BLOCK_SIZE >= 8192:
        num_warps = 16
    elif BLOCK_SIZE >= 2048:
        num_warps = 8
    return BLOCK_SIZE, num_warps


def compare_version(package: str, operator: Callable, target: str):
    try:
        pkg = importlib.import_module(package)
    except ImportError:
        return False
    pkg_version = Version(pkg.__version__)
    return operator(pkg_version, Version(target))



from abc import ABC

import einops
import numpy as np
import torch


class Normalize(torch.nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.dim = dim

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.nn.functional.normalize(x, dim=self.dim, p=2)


class LearnableLogitScaling(torch.nn.Module):
    def __init__(
        self,
        logit_scale_init: float = 1 / 0.07,
        learnable: bool = True,
        max_logit_scale: float = 100,
    ) -> None:
        super().__init__()
        self.max_logit_scale = max_logit_scale
        self.logit_scale_init = logit_scale_init
        self.learnable = learnable
        log_logit_scale = torch.ones([]) * np.log(self.logit_scale_init)
        if learnable:
            self.log_logit_scale = torch.nn.Parameter(log_logit_scale)
        else:
            self.register_buffer('log_logit_scale', log_logit_scale)

    def forward(self, x: torch.Tensor):
        return torch.clip(self.log_logit_scale.exp(), max=self.max_logit_scale) * x

    def extra_repr(self):
        st = (
            f'logit_scale_init={self.logit_scale_init},learnable={self.learnable},'
            f' max_logit_scale={self.max_logit_scale}'
        )
        return st


class EinOpsRearrange(torch.nn.Module):
    def __init__(self, rearrange_expr: str) -> None:
        super().__init__()
        self.rearrange_expr = rearrange_expr

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return einops.rearrange(x, self.rearrange_expr)


class VerboseNNModule(torch.nn.Module, ABC):
    """
    Wrapper around nn.Module that prints registered buffers and parameter names.
    """

    @staticmethod
    def get_readable_tensor_repr(name: str, t: torch.Tensor) -> str:
        return f'({name}): tensor({t.shape}, requires_grad={t.requires_grad})\n'

    def extra_repr(self) -> str:
        named_modules = list(set(p[0] for p in self.named_modules()))

        string_repr: str = ''
        for name, value in self.named_parameters():
            if name.split('.')[0] not in named_modules:
                string_repr += self.get_readable_tensor_repr(name, value)

        for name, value in self.named_buffers():
            string_repr += self.get_readable_tensor_repr(name.split('.')[0], value)

        return string_repr


def cast_if_src_dtype(
    tensor: torch.Tensor, src_dtype: torch.dtype, tgt_dtype: torch.dtype
) -> tuple[torch.Tensor, bool]:
    updated = False
    if tensor.dtype == src_dtype:
        tensor = tensor.to(dtype=tgt_dtype)
        updated = True
    return tensor, updated


class SelectElement(torch.nn.Module):
    def __init__(self, index: int) -> None:
        super().__init__()
        self.index = index

    def forward(self, x: torch.Tensor):
        assert x.ndim >= 3
        return x[:, self.index, ...]


class SelectEOSAndProject(torch.nn.Module):
    """
    Text Pooling used in OpenCLIP
    """

    def __init__(self, proj: torch.nn.Module) -> None:
        super().__init__()
        self.proj = proj

    def forward(self, x, seq_len):
        assert x.ndim == 3
        # x is of shape B x L x D
        # take features from the eot embedding (eot_token is the highest number in each sequence)
        x = x[torch.arange(x.shape[0]), seq_len]
        x = self.proj(x)
        return x



import random
from abc import ABC
from itertools import accumulate
from pathlib import Path
from typing import Any, overload

import numpy as np
import numpy.typing as npt
import torch
from transformers import PreTrainedTokenizerBase

from turbo_alignment.common.data.io import read_jsonl
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.constants import DISABLE_LOSS_LABEL
from turbo_alignment.dataset.base import AlignmentDataset
from turbo_alignment.dataset.chat.models import (
    ChatDatasetRecord,
    ChatMessage,
    ChatMessageRole,
)
from turbo_alignment.dataset.registry import ChatDatasetTypeRegistry
from turbo_alignment.settings.datasets.base import (
    DatasetSourceSettings,
    DatasetStrategy,
)
from turbo_alignment.settings.datasets.chat import ChatDatasetSettings

from .conversation import Conversation

logger = get_project_logger()


class ChatDataset(AlignmentDataset[ChatDatasetRecord], ABC):
    def __init__(
        self,
        source: DatasetSourceSettings,
        settings: ChatDatasetSettings,
        tokenizer: PreTrainedTokenizerBase,
        read: bool = True,
    ) -> None:
        super().__init__(source=source, settings=settings, tokenizer=tokenizer)
        self.settings: ChatDatasetSettings = settings

        if read:
            self._read()

    @overload
    def __tokenize(self, inputs: str) -> npt.NDArray[np.int64]:
        ...

    @overload
    def __tokenize(self, inputs: list[str]) -> np.ndarray:
        ...

    def __tokenize(self, inputs):
        return self.tokenizer(
            inputs,
            add_special_tokens=False,
            padding=False,
            truncation=False,
            return_tensors='np',
        )['input_ids']

    def __keep_end(
        self,
        conversation: Conversation,
        replicas_cum_len: list[int],
        inference: bool,
        max_tokens: int | None,
    ) -> tuple[int, int]:
        if max_tokens is None:
            return 0, len(replicas_cum_len)

        _, right_bound = self.__keep_start(
            conversation=conversation,
            replicas_cum_len=replicas_cum_len,
            inference=inference,
        )

        left_bound = 0
        last_index = replicas_cum_len[-1]
        for i, replica_end_index in enumerate(replicas_cum_len):
            if last_index - replica_end_index <= max_tokens:
                return left_bound, right_bound

            left_bound = i

        raise ValueError('Can\'t trim dialogue to fit all requirements')

    def __keep_start(
        self,
        conversation: Conversation,
        replicas_cum_len: list[int],
        inference: bool,
        max_tokens: int | None = None,
    ) -> tuple[int, int]:
        for i, (message, end_index) in enumerate(zip(conversation.messages[::-1], replicas_cum_len[::-1])):
            if self.settings.only_answer_loss:
                if inference and message.role == ChatMessageRole.BOT:
                    continue
                if not inference and message.role != ChatMessageRole.BOT:
                    continue

            if max_tokens is None or end_index < max_tokens:
                return 0, len(replicas_cum_len) - i

        raise ValueError('Can\'t trim dialogue to fit all requirements')

    def __truncate(
        self,
        conversation: Conversation,
        replicas_cum_len: list[int],
        inference: bool,
        max_tokens: int | None,
    ) -> tuple[int, int]:
        '''
        truncate dialogue to fit all requirements:
        - cumulative tokens num is less than max_tokens
        - keep dialogue end or start according to settings
        - remove bot's messages from dialog end at inference
        - remove user's messages from dialog end at non inference (training)

        returns [first_message_index, last_message_index]
        '''
        if self.settings.keep_end:
            return self.__keep_end(
                conversation=conversation,
                replicas_cum_len=replicas_cum_len,
                max_tokens=max_tokens,
                inference=inference,
            )

        return self.__keep_start(
            conversation=conversation,
            replicas_cum_len=replicas_cum_len,
            max_tokens=max_tokens,
            inference=inference,
        )

    @staticmethod
    def _all_loss_disabled(leftover_messages: list[ChatMessage]) -> bool:
        loss_flags = [message.disable_loss for message in leftover_messages]
        return sum(loss_flags) == len(leftover_messages)

    def _truncate_and_merge(
        self,
        conversation: Conversation,
        tokenized_replicas: list[np.ndarray],
        role_prefix_tokens: dict[ChatMessageRole, np.ndarray],
        suffix_tokens: np.ndarray,
        inference: bool,
        random_cut: bool,
    ) -> tuple[np.ndarray, np.ndarray, str]:
        # random_cut используется только когда inference=true
        assert inference or not random_cut

        bot_prefix_tokens = role_prefix_tokens[ChatMessageRole.BOT]
        max_tokens = (
            self.settings.max_tokens_count - (len(bot_prefix_tokens) if inference else 1)
            if self.settings.max_tokens_count is not None
            else None
        )

        replicas_len = [
            len(r) + len(role_prefix_tokens[m.role]) + len(suffix_tokens)
            for r, m in zip(tokenized_replicas, conversation.messages)
        ]

        left_bound, right_bound = self.__truncate(
            conversation=conversation,
            replicas_cum_len=list(accumulate(replicas_len)),
            max_tokens=max_tokens,
            inference=inference,
        )

        if not inference and right_bound - left_bound < 2:
            raise ValueError('Less than two messages left after truncation')
        if (
            inference
            and left_bound == 0
            and right_bound == 1  # если при инференсе остался только системный промпт
            and conversation.messages[0].role == ChatMessageRole.SYSTEM
        ):
            raise ValueError('Less than two messages left after truncation')
        if not inference and self._all_loss_disabled(conversation.messages[left_bound:right_bound]):
            raise ValueError('No messages with enabled loss were left after truncations')

        if random_cut:
            bot_indices = [
                i
                for i, m in enumerate(conversation.messages)
                if m.role == ChatMessageRole.BOT and left_bound <= i < right_bound
            ]
            right_bound = random.choice(bot_indices) if bot_indices else right_bound

        input_ids = np.array([])
        labels = np.array([])

        truncated_conversation_messages = conversation.messages[left_bound:right_bound]
        truncated_tokenized_replicas = tokenized_replicas[left_bound:right_bound]

        if self.source.system_prompt is not None and self.settings.keep_end and left_bound != 0:
            truncated_conversation_messages = [conversation.messages[0]] + truncated_conversation_messages
            truncated_tokenized_replicas = [truncated_tokenized_replicas[0]] + truncated_tokenized_replicas

        for ind, (message, tokenized_replica) in enumerate(
            zip(
                truncated_conversation_messages,
                truncated_tokenized_replicas,
            )
        ):
            prefix_tokens = role_prefix_tokens[message.role]
            merged_replica = np.concatenate((prefix_tokens, tokenized_replica, suffix_tokens))
            input_ids = np.concatenate((input_ids, merged_replica))

            if (
                (self.settings.only_last_replica_loss and ind != right_bound - left_bound - 1)
                or (self.settings.only_answer_loss and message.role != ChatMessageRole.BOT)
                or message.disable_loss
            ):
                replica_labels = np.full(merged_replica.shape, DISABLE_LOSS_LABEL)
            else:
                replica_labels = np.concatenate(
                    (
                        np.full(prefix_tokens.shape, DISABLE_LOSS_LABEL),
                        tokenized_replica,
                        suffix_tokens,
                    )
                )
            labels = np.concatenate((labels, replica_labels))

        if inference:
            input_ids = np.concatenate((input_ids, bot_prefix_tokens))
            labels = np.concatenate((labels, np.full(bot_prefix_tokens.shape, DISABLE_LOSS_LABEL)))
        else:
            input_ids = np.append(input_ids, self.tokenizer.eos_token_id)
            labels = np.append(labels, DISABLE_LOSS_LABEL)

        # FIXME
        start_replica_token_id = role_prefix_tokens[ChatMessageRole.BOT][0].item()

        # -1 for bos token
        input_ids = input_ids[-(self.settings.max_tokens_count - 1) :]  # type: ignore[operator]
        replica_start_token_inds = np.where(input_ids == start_replica_token_id)[0]
        if len(replica_start_token_inds) != 0:
            cut_index = replica_start_token_inds[0]
            input_ids = input_ids[cut_index:]

        labels = labels[-len(input_ids) :]
        input_ids = np.concatenate((np.array([self.tokenizer.bos_token_id]), input_ids))
        labels = np.concatenate((np.array([DISABLE_LOSS_LABEL]), labels))

        return input_ids, labels, conversation.get_prompt_repr(left_bound, right_bound)

    def _encode(
        self,
        records: list[ChatDatasetRecord],
        inference: bool,
        random_cut: bool,
    ) -> list[dict[str, Any] | None]:
        conversations = [
            Conversation(
                system_prompt=self.source.system_prompt,
                messages=r.messages,
                ignore_system_prompt=self.settings.ignore_system_prompt,
            )
            for r in records
        ]

        logger.info(f'Tokenizing dataset {self.source.name}')
        tokenized_replicas = self.__tokenize([m.content for c in conversations for m in c.messages])

        tokenized_conversations = []
        offset = 0
        for c in conversations:
            tokenized_conversations.append([tokenized_replicas[offset + i] for i in range(len(c.messages))])
            offset += len(c.messages)

        role_prefix_tokens = {
            role: self.__tokenize(
                self.settings.prompt_template.prefix_template.format(
                    role=self.settings.prompt_template.role_tag_mapping[role]
                )
            )[0]
            for role in ChatMessageRole
        }
        # TODO: what if suffix is empty?
        suffix_tokens = self.__tokenize(self.settings.prompt_template.suffix_template)[0]

        logger.info(f'Postprocessing tokenized data in {self.source.name}')
        output: list[dict[str, Any] | None] = []
        for record, conversation, tokenized_replicas in zip(records, conversations, tokenized_conversations):
            try:
                input_ids, labels, prompt = self._truncate_and_merge(
                    conversation=conversation,
                    tokenized_replicas=tokenized_replicas,
                    role_prefix_tokens=role_prefix_tokens,
                    suffix_tokens=suffix_tokens,
                    inference=inference,
                    random_cut=random_cut,
                )

            except ValueError as ex:
                output.append(None)
                logger.warning(f'Sample dropped: {ex}')
                continue

            encoded_record: dict[str, Any] = {
                # 'id': record.id, FIXME: dont work with collators
                'input_ids': torch.LongTensor(input_ids),
                'labels': torch.LongTensor(labels),
                'attention_mask': torch.ones(input_ids.shape, dtype=torch.int64),
            }
            if inference:
                encoded_record.update(
                    {'prompt': prompt, 'id': record.id, 'messages': record.messages, 'meta': record.meta}
                )

            output.append(encoded_record)

        return output

    @staticmethod
    @overload
    def _read_records(records: Path) -> list[ChatDatasetRecord]:
        ...

    @staticmethod
    @overload
    def _read_records(records: list[dict]) -> list[ChatDatasetRecord]:
        ...

    @staticmethod
    def _read_records(records) -> list[ChatDatasetRecord]:
        if isinstance(records, Path):
            return [ChatDatasetRecord(**record) for record in read_jsonl(records)]
        if isinstance(records, list):
            return [ChatDatasetRecord(**record) for record in records]
        raise NotImplementedError


@ChatDatasetTypeRegistry.register(DatasetStrategy.TRAIN)
class TrainChatDataset(ChatDataset):
    def convert_records(self, records: list[ChatDatasetRecord]) -> list[dict[str, Any] | None]:
        return self._encode(records, inference=False, random_cut=False)


@ChatDatasetTypeRegistry.register(DatasetStrategy.INFERENCE)
class InferenceChatDataset(ChatDataset):
    def __init__(
        self,
        source: DatasetSourceSettings,
        settings: ChatDatasetSettings,
        tokenizer: PreTrainedTokenizerBase,
        read: bool = True,
        random_cut: bool = False,
    ) -> None:
        self._random_cut = random_cut

        super().__init__(source=source, settings=settings, tokenizer=tokenizer, read=read)

    def convert_records(self, records: list[ChatDatasetRecord]) -> list[dict[str, Any] | None]:
        return self._encode(records, inference=True, random_cut=self._random_cut)



from abc import abstractmethod
from typing import Generic, TypeVar

from accelerate import Accelerator
from accelerate.utils import set_seed
from accelerate.utils.operations import gather_object
from transformers import PreTrainedModel, PreTrainedTokenizerBase

from turbo_alignment.common.tf.loaders.model.model import load_model
from turbo_alignment.common.tf.loaders.tokenizer import load_tokenizer
from turbo_alignment.dataset.sampling.models import SamplingDatasetRecord
from turbo_alignment.dataset.sampling.rm import SamplingRMDataset
from turbo_alignment.generators.rm import RMSamplingGenerator
from turbo_alignment.pipelines.sampling.base import BaseSamplingStrategy
from turbo_alignment.settings.generators.outputs.rm import RMSamplingInferenceOutput
from turbo_alignment.settings.pipelines.sampling import (
    BaseSamplingWithRMSettings,
    SamplingWithRMSettings,
)

SamplingSettingsWithRMT = TypeVar('SamplingSettingsWithRMT', bound=BaseSamplingWithRMSettings)


class BaseSamplingStrategyWithRM(BaseSamplingStrategy[SamplingSettingsWithRMT], Generic[SamplingSettingsWithRMT]):
    @staticmethod
    def _get_rewards(
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        dataset: SamplingRMDataset,
        experiment_settings: SamplingSettingsWithRMT,
        accelerator: Accelerator,
    ) -> dict[str, dict[str, float]]:
        generator = RMSamplingGenerator(
            model=model,
            tokenizer=tokenizer,
            accelerator=accelerator,
            batch=experiment_settings.rm_batch_size,
            micro_batch=experiment_settings.rm_batch_size,
        )
        outputs: list[RMSamplingInferenceOutput] = gather_object(generator.generate_from_dataset(dataset))[
            : len(dataset)
        ]
        return {out.id: out.rewards for out in outputs}

    @staticmethod
    @abstractmethod
    def _sample(
        experiment_settings: SamplingSettingsWithRMT,
        dataset: SamplingRMDataset,
        rewards: dict[str, dict[str, float]],
    ) -> list[SamplingDatasetRecord]:
        ...

    def sample(self, experiment_settings: SamplingSettingsWithRMT) -> list[SamplingDatasetRecord]:
        accelerator = Accelerator()
        set_seed(seed=0, device_specific=False)

        tokenizer = load_tokenizer(
            experiment_settings.tokenizer_settings,
            experiment_settings.rm,
        )

        model = load_model(experiment_settings.rm, tokenizer)
        if accelerator is not None:
            model = accelerator.prepare_model(model, device_placement=True, evaluation_mode=True)

        model.eval()

        dataset = SamplingRMDataset(
            source=experiment_settings.dataset_settings.sources[0],
            settings=experiment_settings.dataset_settings.chat_dataset,
            tokenizer=tokenizer,
        )

        rewards = self._get_rewards(
            model=model,
            tokenizer=tokenizer,
            dataset=dataset,
            experiment_settings=experiment_settings,
            accelerator=accelerator,
        )

        return self._sample(experiment_settings, dataset, rewards)


class SamplingStrategyWithRM(BaseSamplingStrategyWithRM[SamplingWithRMSettings]):
    @staticmethod
    def _sample(
        experiment_settings: SamplingWithRMSettings,
        dataset: SamplingRMDataset,
        rewards: dict[str, dict[str, float]],
    ) -> list[SamplingDatasetRecord]:
        records = []

        for record_id, record in dataset.original_records_map.items():
            record.rewards = rewards[record_id]
            records.append(record)

        return records



from transformers import PreTrainedModel
from transformers.tokenization_utils_base import PreTrainedTokenizerBase

from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.settings.tf.special_tokens_setter import SpecialTokensSettings

logger = get_project_logger()


class SpecialTokensSetter:
    def __init__(self, tokenizer: PreTrainedTokenizerBase, special_tokens_settings: SpecialTokensSettings) -> None:
        self._tokenizer = tokenizer
        self._special_tokens_settings = special_tokens_settings
        self._special_tokens_already_set: bool = False

    def setBOS(self, bos_token: str) -> None:
        if self._tokenizer.bos_token_id is None:
            logger.info('Model does not have bos_token_id')
            self._tokenizer.add_special_tokens(special_tokens_dict={'bos_token': bos_token})
            assert self._tokenizer.bos_token_id is not None
            logger.info(f'Created bos_token_id = {self._tokenizer.bos_token_id}')
        else:
            logger.info(f'Model has bos_token_id = {self._tokenizer.bos_token_id}')

    def setEOS(self, eos_token: str) -> None:
        if self._tokenizer.eos_token_id is None:
            logger.info('Model does not have eos_token_id')
            self._tokenizer.add_special_tokens(special_tokens_dict={'eos_token': eos_token})
            assert self._tokenizer.eos_token_id is not None
            logger.info(f'Created eos_token_id = {self._tokenizer.eos_token_id}')
        else:
            logger.info(f'Model has eos_token_id = {self._tokenizer.eos_token_id}')

    def setPAD(self, pad_token: str | None) -> None:
        if self._tokenizer.pad_token_id is None and pad_token is None:
            logger.info('Skip adding pad_token_id')
            return None

        if self._tokenizer.pad_token_id is None:
            logger.info('Model does not have pad_token_id')
            self._tokenizer.add_special_tokens(special_tokens_dict={'pad_token': pad_token})
            assert self._tokenizer.pad_token_id is not None
            logger.info(f'Created pad_token_id = {self._tokenizer.pad_token_id}')
        else:
            logger.info(f'Model has pad_token_id = {self._tokenizer.pad_token_id}')

        return None

    def setUNK(self, unk_token: str | None) -> None:
        if self._tokenizer.unk_token_id is None and unk_token is None:
            logger.info('Skip adding sep_token_id')
            return None

        if self._tokenizer.unk_token_id is None:
            logger.info('Model does not have unk_token_id')
            self._tokenizer.add_special_tokens(special_tokens_dict={'unk_token': unk_token})
            assert self._tokenizer.unk_token_id is not None
            logger.info(f'Created unk_token_id = {self._tokenizer.unk_token_id}')
        else:
            logger.info(f'Model has unk_token_id = {self._tokenizer.unk_token_id}')

        return None

    def setSEP(self, sep_token: str | None) -> None:
        if self._tokenizer.sep_token_id is None and sep_token is None:
            logger.info('Skip adding sep_token_id')
            return None

        if self._tokenizer.sep_token_id is None:
            logger.info('Model does not have sep_token_id')
            self._tokenizer.add_special_tokens(special_tokens_dict={'sep_token': sep_token})
            assert self._tokenizer.sep_token_id is not None
            logger.info(f'Created sep_token_id = {self._tokenizer.sep_token_id}')
        else:
            logger.info(f'Model has sep_token_id = {self._tokenizer.sep_token_id}')

        return None

    def set_all(self) -> None:
        self.setBOS(bos_token=self._special_tokens_settings.bos_token)
        self.setEOS(eos_token=self._special_tokens_settings.eos_token)
        self.setPAD(pad_token=self._special_tokens_settings.pad_token)
        self.setUNK(unk_token=self._special_tokens_settings.unk_token)
        self.setSEP(sep_token=self._special_tokens_settings.sep_token)

    def set_custom_tokens(self, tokens: list[str]) -> None:
        if self._special_tokens_already_set:
            raise ValueError('Additional special tokens already set')

        self._special_tokens_already_set = True

        logger.info(f'Added custom special tokens: {tokens}')
        added_tokens = self._tokenizer.add_special_tokens({'additional_special_tokens': tokens})
        assert added_tokens == len(tokens)

    def setup_model_config(self, model: PreTrainedModel) -> None:
        model.config.bos_token_id = self._tokenizer.bos_token_id
        model.config.eos_token_id = self._tokenizer.eos_token_id

        if self._tokenizer.pad_token_id is not None:
            model.config.pad_token_id = self._tokenizer.pad_token_id
        if self._tokenizer.sep_token_id is not None:
            model.config.sep_token_id = self._tokenizer.sep_token_id



from pathlib import Path

import cv2
import numpy as np
import torch
import torchvision.transforms as transforms
import torchvision.transforms.functional as F

from turbo_alignment.common.data.multimodal.image.base import BaseImageReader
from turbo_alignment.common.data.multimodal.registry import ImageModalityReaderRegistry
from turbo_alignment.settings.modality import ModalityReader


@ImageModalityReaderRegistry.register(ModalityReader.IMAGEBIND)
class ImageBindImageReader(BaseImageReader):
    def __init__(self):
        self.data_transform = transforms.Compose(
            [
                transforms.Resize((224, 224), interpolation=F.InterpolationMode.BICUBIC),
                transforms.Normalize(
                    mean=[0.48145466, 0.4578275, 0.40821073], std=[0.26862954, 0.26130358, 0.27577711]
                ),
            ]
        )

    def read(self, path: str) -> torch.Tensor:
        image = cv2.imread(path)
        if image is None:
            raise OSError(f'Image not found: {path}')

        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        transformed_data = self.data_transform(image=image)

        return torch.Tensor(np.moveaxis(transformed_data['image'], -1, 0))



from functools import partial

import numpy as np
import torch
import torch.nn.functional as F
from sklearn.metrics import (
    accuracy_score,
    balanced_accuracy_score,
    f1_score,
    precision_score,
    recall_score,
    roc_auc_score,
)
from sklearn.utils import compute_class_weight
from torch.utils.data import Dataset
from transformers import EvalPrediction

from turbo_alignment.settings.pipelines.train.classification import (
    ClassificationLossSettings,
)
from turbo_alignment.trainers.custom_loss import CustomLossTrainer


def compute_clf_metrics(eval_pred: EvalPrediction) -> dict[str, float]:
    pred_scores, labels = eval_pred
    predictions = np.argmax(pred_scores, axis=1)
    accuracy = accuracy_score(labels, predictions)
    balanced_accuracy = balanced_accuracy_score(labels, predictions)
    f_score = f1_score(labels, predictions, average='weighted')
    precision = precision_score(labels, predictions, average='weighted')
    recall = recall_score(labels, predictions, average='weighted')
    roc_auc = roc_auc_score(labels, pred_scores[:, 1], average='macro', multi_class='ovo')
    metrics = {
        'accuracy': accuracy,
        'balanced_accuracy': balanced_accuracy,
        'f1-score': f_score,
        'recall': recall,
        'precision': precision,
        'roc_auc': roc_auc,
    }
    return metrics


def classification_loss(
    logits: torch.Tensor, labels: torch.LongTensor, loss_settings: ClassificationLossSettings
) -> torch.Tensor:
    if loss_settings.alpha is None:
        alpha = torch.ones((logits.size(-1),), device=logits.device, dtype=logits.dtype)
    else:
        alpha = torch.tensor(loss_settings.alpha, device=logits.device, dtype=logits.dtype)

    ce_loss = F.cross_entropy(logits, labels, weight=alpha, reduction='none')

    p_t = torch.exp(-ce_loss)

    focal_loss = ((1 - p_t) ** loss_settings.gamma) * ce_loss

    return focal_loss.mean()


def auto_class_weights(dataset: Dataset) -> list[float]:
    labels = [dataset[i]['labels'] for i in range(len(dataset))]  # type: ignore[arg-type]
    class_weights = compute_class_weight('balanced', classes=np.unique(labels), y=np.array(labels))
    return class_weights.tolist()


class ClassificationTrainer(CustomLossTrainer):
    def __init__(self, loss_settings: ClassificationLossSettings, **kwargs):
        super().__init__(
            custom_loss=partial(classification_loss, loss_settings=loss_settings),
            compute_metrics=compute_clf_metrics,
            **kwargs,
        )



import os
from abc import abstractmethod
from pathlib import Path
from typing import Callable, Generic, TypeVar

import torch
from torch.utils.data import ConcatDataset, Dataset
from transformers import (
    PreTrainedModel,
    PreTrainedTokenizerBase,
    Trainer,
    TrainingArguments,
)

from turbo_alignment.cherry_picks.base import CherryPickCallbackBase
from turbo_alignment.common.data.io import write_json
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.common.tf.loaders.model import load_model
from turbo_alignment.common.tf.loaders.tokenizer import load_tokenizer
from turbo_alignment.common.tf.special_tokens_setter import SpecialTokensSetter
from turbo_alignment.dataset.loader import DatasetLoader
from turbo_alignment.pipelines.base import BaseStrategy
from turbo_alignment.pipelines.mixin import S3Mixin
from turbo_alignment.pipelines.mixin.logging import LoggingRegistry
from turbo_alignment.settings.datasets.base import DatasetStrategy
from turbo_alignment.settings.pipelines.train.base import BaseTrainExperimentSettings
from turbo_alignment.settings.s3 import ExperimentMetadata, S3HandlerParameters

logger = get_project_logger()


ExperimentSettingsT = TypeVar('ExperimentSettingsT', bound=BaseTrainExperimentSettings)


class BaseTrainStrategy(S3Mixin, BaseStrategy, Generic[ExperimentSettingsT]):
    tokenizer: PreTrainedTokenizerBase
    model: PreTrainedModel
    trainer: Trainer

    @staticmethod
    @abstractmethod
    def _get_cherry_pick_callback(
        experiment_settings: ExperimentSettingsT,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> CherryPickCallbackBase | None:
        ...

    @staticmethod
    def _save_experiment_config(
        experiment_settings: ExperimentSettingsT, model: PreTrainedModel, output_path: Path
    ) -> None:
        model_config = model.config.to_dict()
        model_config['experiment_config'] = experiment_settings
        write_json(model_config, output_path)

    @staticmethod
    def _save_experiment_metadata(experiment_metadata: ExperimentMetadata, output_path: Path) -> None:
        write_json(experiment_metadata.dict(), output_path)

    @staticmethod
    @abstractmethod
    def _get_data_collator(
        experiment_settings: ExperimentSettingsT, tokenizer: PreTrainedTokenizerBase, **kwargs
    ) -> Callable:
        ...

    @staticmethod
    @abstractmethod
    def _get_trainer(
        training_args: TrainingArguments,
        experiment_settings: ExperimentSettingsT,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        train_dataset: Dataset,
        val_dataset: Dataset,
        data_collator: Callable,
    ) -> Trainer:
        ...

    @staticmethod
    def _load_model(
        experiment_settings: ExperimentSettingsT, tokenizer: PreTrainedTokenizerBase
    ) -> torch.nn.Module | PreTrainedModel:
        return load_model(experiment_settings.model_settings, tokenizer)

    @staticmethod
    @abstractmethod
    def _get_training_args(experiment_settings: ExperimentSettingsT) -> TrainingArguments:
        ...

    @staticmethod
    def _load_tokenizer(experiment_settings: ExperimentSettingsT) -> PreTrainedTokenizerBase:
        return load_tokenizer(experiment_settings.tokenizer_settings, experiment_settings.model_settings)

    @abstractmethod
    def _dataset_and_collator_sanity_check(self, dataset: Dataset, collator: Callable) -> None:
        ...

    @staticmethod
    def _get_additional_special_tokens(
        experiment_settings: BaseTrainExperimentSettings,
    ) -> list[str]:
        embeddings_initialization_strategy = experiment_settings.model_settings.embeddings_initialization_strategy
        return list(embeddings_initialization_strategy.keys()) if embeddings_initialization_strategy else []

    def _add_trainer_callbacks(self, experiment_settings: ExperimentSettingsT, **kwargs) -> None:
        if self.trainer.accelerator.is_main_process:
            self.trainer.add_callback(
                LoggingRegistry.by_name(experiment_settings.logging_settings.__name__).get_logging_callback(
                    experiment_settings=experiment_settings
                )
            )

        cherry_pick_callback = self._get_cherry_pick_callback(experiment_settings, self.tokenizer, **kwargs)

        if cherry_pick_callback is not None:
            self.trainer.add_callback(cherry_pick_callback)

        if experiment_settings.checkpoint_uploader_callback_parameters is not None:
            if self.trainer.is_deepspeed_enabled and self.trainer.args.deepspeed_plugin.zero_stage == 3:
                raise NotImplementedError(
                    'You should not use checkpoint uploader callback when DeepSpeed ZeRO stage 3 enabled'
                )
            s3_handler = self._get_s3_handler(S3HandlerParameters())

            self.trainer.add_callback(
                self._get_checkpoint_uploader_callback(
                    s3_handler=s3_handler,
                    checkpoint_uploader_callback_parameters=(
                        experiment_settings.checkpoint_uploader_callback_parameters
                    ),
                )
            )

    def run(self, experiment_settings: ExperimentSettingsT) -> None:
        training_args = self._get_training_args(experiment_settings)

        self.tokenizer = self._load_tokenizer(experiment_settings)

        logger.info('Tokenizer is loaded!')

        additional_special_tokens = self._get_additional_special_tokens(experiment_settings)
        logger.info(f'Special tokens: {additional_special_tokens}')
        special_tokens_setter = SpecialTokensSetter(self.tokenizer, experiment_settings.special_tokens_settings)
        special_tokens_setter.set_all()
        special_tokens_setter.set_custom_tokens(additional_special_tokens)

        logger.info('Special tokens added!')

        self.model = self._load_model(experiment_settings, self.tokenizer)

        special_tokens_setter.setup_model_config(self.model)

        logger.info('Model is loaded!')

        train_dataset: ConcatDataset = ConcatDataset(
            datasets=DatasetLoader().load_datasets(
                experiment_settings.train_dataset_settings,
                tokenizer=self.tokenizer,
                strategy=DatasetStrategy.TRAIN,
            )
        )

        val_dataset: ConcatDataset = ConcatDataset(
            datasets=DatasetLoader().load_datasets(
                experiment_settings.val_dataset_settings,
                tokenizer=self.tokenizer,
                strategy=DatasetStrategy.TRAIN,
            )
        )

        data_collator = self._get_data_collator(experiment_settings, self.tokenizer)

        self.trainer = self._get_trainer(
            training_args,
            experiment_settings,
            self.model,
            self.tokenizer,
            train_dataset,
            val_dataset,
            data_collator,
        )

        if self.trainer.accelerator.is_main_process:
            self._dataset_and_collator_sanity_check(train_dataset, data_collator)

        self._add_trainer_callbacks(experiment_settings)

        os.makedirs(self.trainer.args.output_dir, exist_ok=True)
        self._save_experiment_config(
            experiment_settings, self.trainer.model, Path(self.trainer.args.output_dir) / 'experiment.config'
        )

        experiment_metadata = ExperimentMetadata()
        self._save_experiment_metadata(
            experiment_metadata, Path(self.trainer.args.output_dir) / 'experiment_metadata.json'
        )

        self.trainer.train()

        self.trainer.save_model()



from abc import ABC
from functools import partial

import torch

from turbo_alignment.modeling.imagebind.models import (
    ImageBindArchitectureSettings,
    ModalityType,
)
from turbo_alignment.modeling.imagebind.preprocessors.impl import (
    AudioPreprocessor,
    PadIm2Video,
    PatchEmbedGeneric,
    RGBDTPreprocessor,
    SpatioTemporalPosEmbeddingHelper,
    TextPreprocessor,
)


class Preprocessors(ABC, torch.nn.ModuleDict):
    @staticmethod
    def __get_rgbt_preprocessor(
        video_frames: int,
        kernel_size: tuple[int, int, int],
        stride: tuple[int, int, int],
        vision_embed_dim: int,
    ) -> RGBDTPreprocessor:
        rgbt_stem = PatchEmbedGeneric(
            proj_stem=[
                PadIm2Video(pad_type='repeat', ntimes=2),
                torch.nn.Conv3d(
                    in_channels=3,
                    kernel_size=kernel_size,
                    out_channels=vision_embed_dim,
                    stride=stride,
                    bias=False,
                ),
            ]
        )

        return RGBDTPreprocessor(
            img_size=[3, video_frames, 224, 224],
            num_cls_tokens=1,
            pos_embed_fn=partial(SpatioTemporalPosEmbeddingHelper, learnable=True),
            rgbt_stem=rgbt_stem,
            depth_stem=None,
        )

    @staticmethod
    def __get_text_preprocessor(
        text_embed_dim: int,
    ) -> TextPreprocessor:
        return TextPreprocessor(
            context_length=77,
            vocab_size=49408,
            embed_dim=text_embed_dim,
            causal_masking=True,
        )

    @staticmethod
    def __get_audio_preprocessor(
        audio_kernel_size: int,
        audio_stride: int,
        audio_embed_dim: int,
        audio_num_mel_bins: int,
        audio_target_len: int,
    ):
        audio_stem = PatchEmbedGeneric(
            proj_stem=[
                torch.nn.Conv2d(
                    in_channels=1,
                    kernel_size=audio_kernel_size,
                    stride=audio_stride,
                    out_channels=audio_embed_dim,
                    bias=False,
                ),
            ],
            norm_layer=torch.nn.LayerNorm(normalized_shape=audio_embed_dim),
        )

        return AudioPreprocessor(
            img_size=[1, audio_num_mel_bins, audio_target_len],
            num_cls_tokens=1,
            pos_embed_fn=partial(SpatioTemporalPosEmbeddingHelper, learnable=True),
            audio_stem=audio_stem,
        )

    def __init__(self, settings: ImageBindArchitectureSettings):
        super().__init__()

        video_frames: int = 2

        self[ModalityType.VISION] = self.__get_rgbt_preprocessor(
            video_frames,
            settings.rgbt_kernel_size,
            settings.rgbt_stride,
            settings.vision_embed_dim,
        )
        self[ModalityType.TEXT] = self.__get_text_preprocessor(settings.text_embed_dim)
        self[ModalityType.AUDIO] = self.__get_audio_preprocessor(
            settings.audio_kernel_size,
            settings.audio_stride,
            settings.audio_embed_dim,
            settings.audio_num_mel_bins,
            settings.audio_target_len,
        )



from typing import Iterable

from accelerate import Accelerator
from transformers import PreTrainedModel, PreTrainedTokenizerBase

from turbo_alignment.cherry_picks.base import CherryPickCallbackBase
from turbo_alignment.dataset.chat.conversation import Conversation
from turbo_alignment.dataset.classification.classification import ClassificationDataset
from turbo_alignment.generators.classification import ClassificationGenerator
from turbo_alignment.metrics.metric import Metric
from turbo_alignment.settings.cherry_pick import ClassificationCherryPickSettings
from turbo_alignment.settings.metric import ElementWiseScores, MetricResults


class ClassificationCherryPickCallback(CherryPickCallbackBase[ClassificationDataset]):
    def __init__(
        self,
        cherry_pick_settings: ClassificationCherryPickSettings,
        datasets: Iterable[ClassificationDataset],
        metrics: list[Metric],
    ) -> None:
        super().__init__(cherry_pick_settings=cherry_pick_settings, datasets=datasets, metrics=metrics)

    def _get_dataset_metrics(
        self,
        dataset: ClassificationDataset,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> list[MetricResults]:
        accelerator: Accelerator = kwargs.get('accelerator', None)

        generator = ClassificationGenerator(
            model=model,
            tokenizer=tokenizer,
            accelerator=accelerator,
        )

        generations = generator.generate_from_dataset(dataset)
        predictions = [record.predicted_label for record in generations]
        labels = [record['labels'] for record in dataset]

        contexts = [
            Conversation(
                system_prompt=dataset.source.system_prompt,
                messages=record.messages,
                ignore_system_prompt=dataset.settings.chat_settings.ignore_system_prompt,
            ).get_prompt_repr(0, len(record.messages))
            for record in generations
        ]

        metric_outputs = [
            MetricResults(
                element_wise_scores=[ElementWiseScores(label=dataset.source.name + '@@' + 'contexts', values=contexts)]
            ),
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(label=dataset.source.name + '@@' + 'predictions', values=predictions)
                ]
            ),
            MetricResults(
                element_wise_scores=[ElementWiseScores(label=dataset.source.name + '@@' + 'labels', values=labels)]
            ),
        ]

        return metric_outputs



from pathlib import Path
from typing import Annotated, Sequence

from pydantic import Field

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel
from turbo_alignment.settings.datasets.chat import ChatMultiDatasetSettings
from turbo_alignment.settings.datasets.classification import (
    ClassificationMultiDatasetSettings,
)
from turbo_alignment.settings.datasets.multimodal import MultimodalMultiDatasetSettings
from turbo_alignment.settings.model import (
    PreTrainedAdaptersModelSettings,
    PreTrainedModelSettings,
)
from turbo_alignment.settings.tf.tokenizer import TokenizerSettings

INFERENCE_DATASETS_SETTINGS = Annotated[
    ChatMultiDatasetSettings | ClassificationMultiDatasetSettings | MultimodalMultiDatasetSettings,
    Field(discriminator='dataset_type'),
]


class SingleModelInferenceSettings(ExtraFieldsNotAllowedBaseModel):
    model_settings: PreTrainedAdaptersModelSettings | PreTrainedModelSettings
    tokenizer_settings: TokenizerSettings
    use_vllm: bool = False
    batch: int = 1
    micro_batch: int = 1


class InferenceExperimentSettings(ExtraFieldsNotAllowedBaseModel):
    inference_settings: Sequence[SingleModelInferenceSettings]

    dataset_settings: INFERENCE_DATASETS_SETTINGS
    save_path: Path



from typing import Any

from turbo_alignment.dataset.pair_preferences.collators import (
    PairPreferenceDataCollator,
)


class KTODataCollator(PairPreferenceDataCollator):
    def __call__(self, examples: list[dict[str, dict[str, Any]]]) -> dict[str, Any]:
        max_length = 0
        for ex in examples:
            max_length = max(max_length, len(ex['chat']['input_ids']))
            max_length = max(max_length, len(ex['KL']['input_ids']))

        chat_batch = dict(self._get_batch(examples, self.tokenizer, 'chat', max_length))
        kl_batch = dict(self._get_batch(examples, self.tokenizer, 'KL', max_length))
        desirable_batch = [ex['is_desirable'] for ex in examples]

        return {
            'is_desirable': desirable_batch,
            'input_ids': chat_batch['input_ids'],
            'attention_mask': chat_batch['attention_mask'],
            'labels': chat_batch['labels'],
            'KL_input_ids': kl_batch['input_ids'],
            'KL_attention_mask': kl_batch['attention_mask'],
            'KL_labels': kl_batch['labels'],
        }



from .chat import ChatInferenceStrategy
from .classification import ClassificationInferenceStrategy
from .multimodal import MultimodalInferenceStrategy
from .rag import RAGInferenceStrategy
from .rm import RMInferenceStrategy



from pydantic import Field

from turbo_alignment.settings.cherry_pick import ChatCherryPickSettings
from turbo_alignment.settings.datasets.kto import KTOMultiDatasetSettings
from turbo_alignment.settings.pipelines.train.base import BaseTrainExperimentSettings
from turbo_alignment.settings.pipelines.train.dpo import SyncRefModelSettings
from turbo_alignment.settings.tf.trainer import TrainerSettings


class KTOTrainerSettings(TrainerSettings):
    undesirable_weight: float = 1.0
    desirable_weight: float = 1.33
    beta: float = 0.1
    use_ref_model: bool = True
    sync_ref_settings: SyncRefModelSettings = SyncRefModelSettings()
    average_log_prob: bool = Field(default=False, description='Normalize log probability by length or not')


class KTOTrainExperimentSettings(BaseTrainExperimentSettings):
    train_dataset_settings: KTOMultiDatasetSettings
    val_dataset_settings: KTOMultiDatasetSettings

    cherry_pick_settings: ChatCherryPickSettings

    trainer_settings: KTOTrainerSettings



from dataclasses import dataclass

import torch
from transformers.modeling_outputs import ModelOutput


@dataclass
class RetrieveAugLMOutput(ModelOutput):
    # RetrieveAugLMOutput should not have more than one required field.
    logits: torch.FloatTensor
    labels: torch.Tensor | None = None
    doc_scores: torch.Tensor | None = None
    past_key_values: list[torch.FloatTensor] | None = None
    retrieved_doc_embeds: torch.FloatTensor | None = None
    retrieved_doc_ids: torch.LongTensor | None = None
    joined_input_ids: torch.Tensor | None = None
    joined_attention_mask: torch.Tensor | None = None
    question_encoder_last_hidden_state: torch.FloatTensor | None = None
    question_enc_hidden_states: tuple[torch.FloatTensor] | None = None
    question_enc_attentions: tuple[torch.FloatTensor] | None = None
    generator_enc_last_hidden_state: torch.FloatTensor | None = None
    generator_enc_hidden_states: tuple[torch.FloatTensor] | None = None
    generator_enc_attentions: tuple[torch.FloatTensor] | None = None
    generator_dec_hidden_states: tuple[torch.FloatTensor] | None = None
    generator_dec_attentions: tuple[torch.FloatTensor] | None = None
    generator_cross_attentions: tuple[torch.FloatTensor] | None = None


@dataclass
class RetrieveAugLMMarginOutput(ModelOutput):
    # RetrieveAugLMMarginOutput should not have more than one required field.
    loss: torch.Tensor
    logits: torch.FloatTensor | None = None
    doc_scores: torch.Tensor | None = None
    past_key_values: list[torch.FloatTensor] | None = None
    retrieved_doc_embeds: torch.FloatTensor | None = None
    retrieved_doc_ids: torch.LongTensor | None = None
    context_input_ids: torch.LongTensor | None = None
    context_attention_mask: torch.LongTensor | None = None
    question_encoder_last_hidden_state: torch.FloatTensor | None = None
    question_enc_hidden_states: tuple[torch.FloatTensor] | None = None
    question_enc_attentions: tuple[torch.FloatTensor] | None = None
    generator_enc_last_hidden_state: torch.FloatTensor | None = None
    generator_enc_hidden_states: tuple[torch.FloatTensor] | None = None
    generator_enc_attentions: tuple[torch.FloatTensor] | None = None
    generator_dec_hidden_states: tuple[torch.FloatTensor] | None = None
    generator_dec_attentions: tuple[torch.FloatTensor] | None = None
    generator_cross_attentions: tuple[torch.FloatTensor] | None = None



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Callable, Literal

import torch
import torch.nn.functional as F
from torch import nn
from torch.utils.data import Dataset
from transformers import (
    DefaultFlowCallback,
    PreTrainedModel,
    PreTrainedTokenizerBase,
    PrinterCallback,
    ProgressCallback,
    Trainer,
    TrainerCallback,
    TrainerControl,
    TrainingArguments,
)
from transformers.integrations import get_reporting_integration_callbacks

from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.common.tf.callbacks.common import MetricsCallbackHandler
from turbo_alignment.common.tf.callbacks.sync_ref_model import SyncRefModelCallback
from turbo_alignment.constants import DISABLE_LOSS_LABEL
from turbo_alignment.settings.pipelines.train.dpo import (
    APODownLossSettings,
    APOZeroLossSettings,
    CPOLossSettings,
    DPOLossesType,
    ASFTLossSettings,
    HingeLossSettings,
    IPOLossSettings,
    KTOLossSettings,
    ORPOLossSettings,
    SigmoidLossSettings,
    SigmoidLossWithMarginSettings,
    SimPOLossSettings,
    SlicHfLossSettings,
    SyncRefModelSettings,
)
from turbo_alignment.trainers.utils import (
    DPOLossRegistry,
    concatenated_inputs,
    prepare_model,
)

logger = get_project_logger()


@DPOLossRegistry.register(DPOLossesType.SIGMOID)
class SigmoidLoss(DPOLossRegistry):
    def __init__(self, *args, beta: float = 0.1, label_smoothing: float = 0, **kwargs) -> None:
        self.beta = beta
        self.label_smoothing = label_smoothing
        super().__init__(*args, **kwargs)

    def compute_loss(
        self,
        policy_chosen_logps: torch.FloatTensor,
        policy_rejected_logps: torch.FloatTensor,
        reference_chosen_logps: torch.FloatTensor,
        reference_rejected_logps: torch.FloatTensor,
        precomputed_margins: torch.FloatTensor | None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        pi_logratios = policy_chosen_logps - policy_rejected_logps
        ref_logratios = reference_chosen_logps - reference_rejected_logps

        logits = pi_logratios - ref_logratios

        chosen_rewards = self.beta * (policy_chosen_logps - reference_chosen_logps).detach()
        rejected_rewards = self.beta * (policy_rejected_logps - reference_rejected_logps).detach()

        loss = (
            -F.logsigmoid(self.beta * logits) * (1 - self.label_smoothing)
            - F.logsigmoid(-self.beta * logits) * self.label_smoothing
        )

        return (
            loss,
            chosen_rewards,
            rejected_rewards,
        )


@DPOLossRegistry.register(DPOLossesType.KTO)
class KTOLoss(DPOLossRegistry):
    def __init__(self, *args, beta: float = 0.1, **kwargs) -> None:
        self.beta = beta
        super().__init__(*args, **kwargs)

    def compute_loss(
        self,
        policy_chosen_logps: torch.FloatTensor,
        policy_rejected_logps: torch.FloatTensor,
        reference_chosen_logps: torch.FloatTensor,
        reference_rejected_logps: torch.FloatTensor,
        precomputed_margins: torch.FloatTensor | None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        chosen_KL = (policy_chosen_logps - reference_chosen_logps).mean().clamp(min=0)
        rejected_KL = (policy_rejected_logps - reference_rejected_logps).mean().clamp(min=0)

        chosen_logratios = policy_chosen_logps - reference_chosen_logps
        rejected_logratios = policy_rejected_logps - reference_rejected_logps

        chosen_rewards = self.beta * (policy_chosen_logps - reference_chosen_logps).detach()
        rejected_rewards = self.beta * (policy_rejected_logps - reference_rejected_logps).detach()

        loss = torch.cat(
            (
                1 - F.sigmoid(self.beta * (chosen_logratios - rejected_KL)),
                1 - F.sigmoid(self.beta * (chosen_KL - rejected_logratios)),
            ),
            0,
        )

        return loss, chosen_rewards, rejected_rewards


@DPOLossRegistry.register(DPOLossesType.IPO)
class IPOLoss(DPOLossRegistry):
    def __init__(self, *args, beta: float = 0.1, **kwargs) -> None:
        self.beta = beta
        super().__init__(*args, **kwargs)

    def compute_loss(
        self,
        policy_chosen_logps: torch.FloatTensor,
        policy_rejected_logps: torch.FloatTensor,
        reference_chosen_logps: torch.FloatTensor,
        reference_rejected_logps: torch.FloatTensor,
        precomputed_margins: torch.FloatTensor | None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        pi_logratios = policy_chosen_logps - policy_rejected_logps
        ref_logratios = reference_chosen_logps - reference_rejected_logps

        logits = pi_logratios - ref_logratios

        chosen_rewards = self.beta * (policy_chosen_logps - reference_chosen_logps).detach()
        rejected_rewards = self.beta * (policy_rejected_logps - reference_rejected_logps).detach()

        loss = (logits - 1 / (2 * self.beta)) ** 2

        return loss, chosen_rewards, rejected_rewards


@DPOLossRegistry.register(DPOLossesType.CPO)
class CPOLoss(DPOLossRegistry):
    def __init__(self, *args, beta: float = 0.1, norm: bool, **kwargs) -> None:
        self.beta = beta
        self.norm = norm
        super().__init__(*args, **kwargs)

    def compute_loss(
        self,
        policy_chosen_logps: torch.FloatTensor,
        policy_rejected_logps: torch.FloatTensor,
        reference_chosen_logps: torch.FloatTensor,
        reference_rejected_logps: torch.FloatTensor,
        precomputed_margins: torch.FloatTensor | None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        pi_logratios = policy_chosen_logps - policy_rejected_logps
        ref_logratios = reference_chosen_logps - reference_rejected_logps

        logits = pi_logratios - ref_logratios

        chosen_rewards = self.beta * (policy_chosen_logps - reference_chosen_logps).detach()
        rejected_rewards = self.beta * (policy_rejected_logps - reference_rejected_logps).detach()

        dpo_loss = -F.logsigmoid(self.beta * logits) if self.norm else -F.logsigmoid(self.beta * pi_logratios)
        sft_loss = -policy_chosen_logps

        loss = dpo_loss + sft_loss

        return loss, chosen_rewards, rejected_rewards


@DPOLossRegistry.register(DPOLossesType.HINGE)
class HingeLoss(DPOLossRegistry):
    def __init__(self, *args, beta: float = 0.1, norm: bool = False, **kwargs) -> None:
        self.beta = beta
        self.norm = norm
        super().__init__(*args, **kwargs)

    def compute_loss(
        self,
        policy_chosen_logps: torch.FloatTensor,
        policy_rejected_logps: torch.FloatTensor,
        reference_chosen_logps: torch.FloatTensor,
        reference_rejected_logps: torch.FloatTensor,
        precomputed_margins: torch.FloatTensor | None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        pi_logratios = policy_chosen_logps - policy_rejected_logps
        ref_logratios = reference_chosen_logps - reference_rejected_logps

        logits = pi_logratios - ref_logratios

        chosen_rewards = self.beta * (policy_chosen_logps - reference_chosen_logps).detach()
        rejected_rewards = self.beta * (policy_rejected_logps - reference_rejected_logps).detach()

        loss = torch.relu(1 - self.beta * (policy_chosen_logps - policy_rejected_logps))

        if self.norm:
            loss = torch.relu(1 - self.beta * logits)

        return loss, chosen_rewards, rejected_rewards


@DPOLossRegistry.register(DPOLossesType.SLIC_HF)
class SlicHfLoss(DPOLossRegistry):
    def __init__(self, delta: float = 1, beta: float = 1.0, lam: float = 1.0, norm: bool = False) -> None:
        self.delta = delta
        self.beta = beta
        self.norm = norm
        self.lam = lam

    def compute_loss(
        self,
        policy_chosen_logps: torch.FloatTensor,
        policy_rejected_logps: torch.FloatTensor,
        reference_chosen_logps: torch.FloatTensor,
        reference_rejected_logps: torch.FloatTensor,
        precomputed_margins: torch.FloatTensor | None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        pi_logratios = policy_chosen_logps - policy_rejected_logps
        ref_logratios = reference_chosen_logps - reference_rejected_logps

        logits = pi_logratios - ref_logratios

        chosen_rewards = self.beta * (policy_chosen_logps - reference_chosen_logps).detach()
        rejected_rewards = self.beta * (policy_rejected_logps - reference_rejected_logps).detach()

        loss = torch.relu(self.delta - self.beta * (policy_chosen_logps - policy_rejected_logps))

        if self.norm:
            loss = torch.relu(self.delta - self.beta * logits)

        if precomputed_margins is not None:
            loss = loss - self.lam * precomputed_margins

        return loss, chosen_rewards, rejected_rewards


@DPOLossRegistry.register(DPOLossesType.SIMPO)
class SimPOLoss(DPOLossRegistry):
    def __init__(self, *args, beta: float = 0.1, gamma: float = 0.1, **kwargs) -> None:
        self.beta = beta
        self.gamma = gamma
        super().__init__(*args, **kwargs)

    def compute_loss(
        self,
        policy_chosen_logps: torch.FloatTensor,
        policy_rejected_logps: torch.FloatTensor,
        reference_chosen_logps: torch.FloatTensor | None,
        reference_rejected_logps: torch.FloatTensor | None,
        precomputed_margins: torch.FloatTensor | None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        pi_logratios = policy_chosen_logps - policy_rejected_logps

        logits = pi_logratios - self.gamma

        chosen_rewards = self.beta * (policy_chosen_logps).detach()
        rejected_rewards = self.beta * (policy_rejected_logps).detach()

        loss = -F.logsigmoid(self.beta * logits)

        return (
            loss,
            chosen_rewards,
            rejected_rewards,
        )


@DPOLossRegistry.register(DPOLossesType.ORPO)
class ORPOLoss(DPOLossRegistry):
    def __init__(self, *args, beta: float = 0.1, **kwargs) -> None:
        self.beta = beta
        super().__init__(*args, **kwargs)

    def compute_loss(
        self,
        policy_chosen_logps: torch.FloatTensor,
        policy_rejected_logps: torch.FloatTensor,
        reference_chosen_logps: torch.FloatTensor | None,
        reference_rejected_logps: torch.FloatTensor | None,
        precomputed_margins: torch.FloatTensor | None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        log_odds = (policy_chosen_logps - policy_rejected_logps) - (
            torch.log1p(-torch.clamp(torch.exp(policy_chosen_logps), max=1 - 1e-7))
            - torch.log1p(-torch.clamp(torch.exp(policy_rejected_logps), max=1 - 1e-7))
        )

        ratio = -F.logsigmoid(log_odds)
        losses = -policy_chosen_logps + self.beta * ratio

        chosen_rewards = self.beta * policy_chosen_logps.detach()
        rejected_rewards = self.beta * policy_rejected_logps.detach()

        return losses, chosen_rewards, rejected_rewards


@DPOLossRegistry.register(DPOLossesType.ASFT)
class ASFTLoss(DPOLossRegistry):
    def __init__(self, *args, beta: float = 0.1, **kwargs) -> None:
        self.beta = beta
        super().__init__(*args, **kwargs)

    def compute_loss(
        self,
        policy_chosen_logps: torch.FloatTensor,
        policy_rejected_logps: torch.FloatTensor,
        reference_chosen_logps: torch.FloatTensor | None,
        reference_rejected_logps: torch.FloatTensor | None,
        precomputed_margins: torch.FloatTensor | None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        chosen_ratio = policy_chosen_logps - (torch.log1p(-torch.clamp(torch.exp(policy_chosen_logps), max=1 - 1e-7)))
        rejected_ratio = policy_rejected_logps - (
            torch.log1p(-torch.clamp(torch.exp(policy_rejected_logps), max=1 - 1e-7))
        )
        sigm_diff = -F.logsigmoid(chosen_ratio) - F.logsigmoid(-rejected_ratio)

        losses = -policy_chosen_logps + self.beta * sigm_diff

        chosen_rewards = self.beta * policy_chosen_logps.detach()
        rejected_rewards = self.beta * policy_rejected_logps.detach()

        return losses, chosen_rewards, rejected_rewards


@DPOLossRegistry.register(DPOLossesType.SIGMOID_WITH_MARGIN)
class SigmoidLossWithMargin(DPOLossRegistry):
    def __init__(self, *args, beta: float = 0.1, **kwargs) -> None:
        self.beta = beta
        super().__init__(*args, **kwargs)

    def compute_loss(
        self,
        policy_chosen_logps: torch.FloatTensor,
        policy_rejected_logps: torch.FloatTensor,
        reference_chosen_logps: torch.FloatTensor,
        reference_rejected_logps: torch.FloatTensor,
        precomputed_margins: torch.FloatTensor | None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        pi_logratios = policy_chosen_logps - policy_rejected_logps
        ref_logratios = reference_chosen_logps - reference_rejected_logps

        logits = pi_logratios - ref_logratios

        chosen_rewards = self.beta * (policy_chosen_logps - reference_chosen_logps).detach()
        rejected_rewards = self.beta * (policy_rejected_logps - reference_rejected_logps).detach()

        if precomputed_margins is None:
            raise ValueError('Precomputed margins should not be none when using SigmoidLossWithMargin')

        loss = -F.logsigmoid(self.beta * logits - precomputed_margins)

        return (
            loss,
            chosen_rewards,
            rejected_rewards,
        )


@DPOLossRegistry.register(DPOLossesType.APO_DOWN)
class APODownLoss(DPOLossRegistry):
    def __init__(self, *args, beta: float = 0.1, **kwargs) -> None:
        self.beta = beta
        super().__init__(*args, **kwargs)

    def compute_loss(
        self,
        policy_chosen_logps: torch.FloatTensor,
        policy_rejected_logps: torch.FloatTensor,
        reference_chosen_logps: torch.FloatTensor,
        reference_rejected_logps: torch.FloatTensor,
        precomputed_margins: torch.FloatTensor | None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        chosen_logratios = policy_chosen_logps - reference_chosen_logps
        rejected_logratios = policy_rejected_logps - reference_rejected_logps

        losses_chosen = F.sigmoid(self.beta * chosen_logratios)
        losses_rejected = 1 - F.sigmoid(self.beta * (chosen_logratios - rejected_logratios))

        loss = losses_chosen + losses_rejected

        chosen_rewards = self.beta * (policy_chosen_logps - reference_chosen_logps).detach()
        rejected_rewards = self.beta * (policy_rejected_logps - reference_rejected_logps).detach()

        return (
            loss,
            chosen_rewards,
            rejected_rewards,
        )


@DPOLossRegistry.register(DPOLossesType.APO_ZERO)
class APOZeroLoss(DPOLossRegistry):
    def __init__(self, *args, beta: float = 0.1, **kwargs) -> None:
        self.beta = beta
        super().__init__(*args, **kwargs)

    def compute_loss(
        self,
        policy_chosen_logps: torch.FloatTensor,
        policy_rejected_logps: torch.FloatTensor,
        reference_chosen_logps: torch.FloatTensor,
        reference_rejected_logps: torch.FloatTensor,
        precomputed_margins: torch.FloatTensor | None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        chosen_logratios = policy_chosen_logps - reference_chosen_logps
        rejected_logratios = policy_rejected_logps - reference_rejected_logps

        losses_chosen = 1 - F.sigmoid(self.beta * chosen_logratios)
        losses_rejected = F.sigmoid(self.beta * rejected_logratios)

        loss = losses_chosen + losses_rejected

        chosen_rewards = self.beta * (policy_chosen_logps - reference_chosen_logps).detach()
        rejected_rewards = self.beta * (policy_rejected_logps - reference_rejected_logps).detach()

        return (
            loss,
            chosen_rewards,
            rejected_rewards,
        )


@dataclass
class DPOTrainingArguments(TrainingArguments):
    loss_settings: (
        SigmoidLossSettings
        | HingeLossSettings
        | IPOLossSettings
        | SlicHfLossSettings
        | KTOLossSettings
        | CPOLossSettings
        | ORPOLossSettings
        | ASFTLossSettings
        | SimPOLossSettings
        | SlicHfLossSettings
        | SigmoidLossWithMarginSettings
        | APOZeroLossSettings
        | APODownLossSettings
    ) = field(
        default_factory=SigmoidLossSettings(loss_type=DPOLossesType.SIGMOID)
    )  # type: ignore[call-overload]
    sync_ref_settings: SyncRefModelSettings = field(  # type: ignore[call-overload]
        default_factory=SyncRefModelSettings()
    )
    use_ref_model: bool = True
    use_sft_model: bool = False
    average_log_prob: bool = False


class DPOTrainer(Trainer):
    """
    Inspired by https://github.com/huggingface/trl/blob/main/trl/trainer/dpo_trainer.py

    """

    def __init__(
        self,
        model: PreTrainedModel | nn.Module,
        data_collator: Callable,
        args: DPOTrainingArguments,
        train_dataset: Dataset,
        eval_dataset: Dataset,
        ref_model: PreTrainedModel | nn.Module | None = None,
        sft_model: PreTrainedModel | nn.Module | None = None,
        tokenizer: PreTrainedTokenizerBase | None = None,
        callbacks: list[TrainerCallback] | None = None,
        **kwargs,
    ):
        self.data_collator = data_collator

        self.average_log_prob = args.average_log_prob
        self.sync_ref_settings = args.sync_ref_settings

        if hasattr(args, 'loss_settings'):
            self.loss_type = args.loss_settings['loss_type']  # type: ignore[index]

            if (
                self.loss_type in (DPOLossesType.SIMPO, DPOLossesType.ORPO, DPOLossesType.ASFT)
                and not args.average_log_prob
            ):
                raise ValueError(f'You should normalize logits by length when using {self.loss_type}')

            loss_args = args.loss_settings
            loss_args.pop('loss_type')  # type: ignore[union-attr]
            self.dpo_loss_registry = DPOLossRegistry.by_name(self.loss_type)(**loss_args)

        self._stored_metrics: dict[str, dict[str, list[float]]] = defaultdict(lambda: defaultdict(list))

        super().__init__(
            model=model,
            args=args,
            data_collator=data_collator,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            tokenizer=tokenizer,
            callbacks=callbacks,
            **kwargs,
        )

        if hasattr(args, 'loss_settings') and self.loss_type in (
            DPOLossesType.SIMPO,
            DPOLossesType.ORPO,
            DPOLossesType.ASFT,
        ):
            logger.info(f'You can turn off ref_model when using {self.loss_type} for memory saving')

        self.ref_model = ref_model
        self.sft_model = sft_model

        if self.ref_model is not None:
            self.ref_model = prepare_model(self.ref_model, self.accelerator, self.is_deepspeed_enabled)

        if self.sft_model is not None:
            self.sft_model = prepare_model(self.sft_model, self.accelerator, self.is_deepspeed_enabled)

        default_callbacks = [DefaultFlowCallback] + get_reporting_integration_callbacks(self.args.report_to)
        callbacks = default_callbacks if callbacks is None else default_callbacks + callbacks
        self.callback_handler = MetricsCallbackHandler(
            callbacks,
            model,
            tokenizer,
            None,
            None,
            ref_model=self.ref_model,
            sft_model=self.sft_model,
            accelerator=self.accelerator,
        )
        self.add_callback(PrinterCallback if self.args.disable_tqdm else ProgressCallback)
        self.control: TrainerControl = self.callback_handler.on_init_end(self.args, self.state, self.control)

        if self.sync_ref_settings['sync_ref_model']:  # type: ignore[index]
            self.add_callback(SyncRefModelCallback(sync_ref_settings=self.sync_ref_settings))

    def dpo_loss(
        self,
        policy_chosen_logps: torch.Tensor,
        policy_rejected_logps: torch.Tensor,
        reference_chosen_logps: torch.Tensor,
        reference_rejected_logps: torch.Tensor,
        precomputed_margins: torch.Tensor | None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        return self.dpo_loss_registry.compute_loss(
            policy_chosen_logps=policy_chosen_logps,
            policy_rejected_logps=policy_rejected_logps,
            reference_chosen_logps=reference_chosen_logps,
            reference_rejected_logps=reference_rejected_logps,
            precomputed_margins=precomputed_margins,
        )

    def _get_batch_logps(
        self,
        logits: torch.Tensor,
        labels: torch.Tensor,
        average_log_prob: bool = False,
    ) -> torch.Tensor:
        if logits.shape[:-1] != labels.shape:
            raise ValueError('Logits (batch and sequence length dim) and labels must have the same shape.')

        labels = labels[:, 1:].clone()
        logits = logits[:, :-1, :]
        loss_mask = labels != DISABLE_LOSS_LABEL

        labels[labels == DISABLE_LOSS_LABEL] = 0

        per_token_logps = torch.gather(logits.log_softmax(-1), dim=2, index=labels.unsqueeze(2)).squeeze(2)

        if average_log_prob:
            return (per_token_logps * loss_mask).sum(-1) / loss_mask.sum(-1)
        return (per_token_logps * loss_mask).sum(-1)

    def concatenated_forward(
        self, model: nn.Module, batch: dict[str, Any]
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor | None]:
        concatenated_batch = concatenated_inputs(batch, device=self.accelerator.device)

        precomputed_margins: torch.Tensor | None = concatenated_batch.pop('margin', None)

        all_logits = model(
            concatenated_batch['input_ids'],
            attention_mask=concatenated_batch['attention_mask'],
        ).logits.to(torch.float32)
        all_logps = self._get_batch_logps(
            all_logits,
            concatenated_batch['labels'],
            average_log_prob=self.average_log_prob,
        )
        chosen_idxs = batch['inputs_w']['input_ids'].shape[0]
        rejected_idx = batch['inputs_l']['input_ids'].shape[0]

        chosen_logps = all_logps[:chosen_idxs]
        rejected_logps = all_logps[chosen_idxs : chosen_idxs + rejected_idx]

        chosen_logits = all_logits[:chosen_idxs]
        rejected_logits = all_logits[chosen_idxs:]
        return chosen_logps, rejected_logps, chosen_logits, rejected_logits, precomputed_margins

    def _get_logps(self, model: nn.Module | None, batch: dict[str, Any]) -> tuple[torch.Tensor, torch.Tensor]:
        with torch.no_grad():
            if model is not None:
                (chosen_logps, rejected_logps, *_) = self.concatenated_forward(model, batch)
            else:
                with self.accelerator.unwrap_model(self.model).disable_adapter():
                    (
                        chosen_logps,
                        rejected_logps,
                        *_,
                    ) = self.concatenated_forward(self.model, batch)

        return chosen_logps, rejected_logps

    def get_batch_metrics(
        self,
        model: nn.Module,
        batch: dict[str, Any],
        train_eval: Literal['train', 'eval'] = 'train',
    ) -> tuple[torch.Tensor, dict[str, float]]:
        metrics: dict[str, float] = {}

        (
            policy_chosen_logps,
            policy_rejected_logps,
            policy_chosen_logits,
            policy_rejected_logits,
            precomputed_margins,
        ) = self.concatenated_forward(model, batch)

        reference_chosen_logps, reference_rejected_logps = torch.Tensor([float('inf')]), torch.Tensor([float('inf')])

        if self.args.use_ref_model or self.loss_type not in (
            DPOLossesType.SIMPO,
            DPOLossesType.ORPO,
            DPOLossesType.ASFT,
        ):
            reference_chosen_logps, reference_rejected_logps = self._get_logps(self.ref_model, batch)

        losses, chosen_rewards, rejected_rewards = self.dpo_loss(
            policy_chosen_logps=policy_chosen_logps,
            policy_rejected_logps=policy_rejected_logps,
            reference_chosen_logps=reference_chosen_logps,
            reference_rejected_logps=reference_rejected_logps,
            precomputed_margins=precomputed_margins,
        )

        prefix = 'eval_' if train_eval == 'eval' else ''

        dpo_prefix_name = prefix + 'rewards/'

        metrics = self._compute_metrics(metrics, dpo_prefix_name, chosen_rewards, rejected_rewards)

        logp_accuracies = (policy_chosen_logps > policy_rejected_logps).float()
        metrics[f'{prefix}logps/accuracies'] = (logp_accuracies).detach().cpu().mean().item()
        metrics[f'{prefix}logps/rejected'] = (policy_rejected_logps).detach().cpu().mean().item()
        metrics[f'{prefix}logps/chosen'] = (policy_chosen_logps).detach().cpu().mean().item()

        metrics[f'{prefix}logits/rejected'] = (policy_rejected_logits).detach().cpu().mean().item()
        metrics[f'{prefix}logits/chosen'] = (policy_chosen_logits).detach().cpu().mean().item()

        if self.args.use_ref_model:
            ref_logp_accuracies = (reference_chosen_logps > reference_rejected_logps).float()
            metrics[f'{prefix}logps/ref_accuracies'] = (ref_logp_accuracies).detach().cpu().mean().item()
            metrics[f'{prefix}logps/ref_rejected'] = (reference_rejected_logps).detach().cpu().mean().item()
            metrics[f'{prefix}logps/ref_chosen'] = (reference_chosen_logps).detach().cpu().mean().item()

            metrics = self._compute_flips(
                metrics, prefix, logp_accuracies.detach().cpu(), ref_logp_accuracies.detach().cpu()
            )

        if self.loss_type == DPOLossesType.KTO:
            kto_chosen_KL = (
                (policy_chosen_logps.detach().cpu() - reference_chosen_logps.detach().cpu()).mean().clamp(min=0)
            )
            kto_rejected_KL = (
                (policy_rejected_logps.detach().cpu() - reference_rejected_logps.detach().cpu()).mean().clamp(min=0)
            )
            kto_z_chosen = self.dpo_loss_registry.beta * (chosen_rewards - kto_chosen_KL)
            kto_z_rejected = self.dpo_loss_registry.beta * (rejected_rewards - kto_rejected_KL)
            kto_grad_term_chosen = (-1 * F.sigmoid(kto_z_chosen) * F.sigmoid(-kto_z_chosen)).mean()
            kto_grad_term_rejected = (1 * F.sigmoid(kto_z_rejected) * F.sigmoid(-kto_z_rejected)).mean()
            kto_grad_term = kto_grad_term_chosen + kto_grad_term_rejected

            metrics[f'{prefix}rewards/kto_grad_term'] = kto_grad_term.item()
            metrics[f'{prefix}rewards/kto_grad_term_chosen'] = kto_grad_term_chosen.item()
            metrics[f'{prefix}rewards/kto_grad_term_rejected'] = kto_grad_term_rejected.item()

        elif self.loss_type == DPOLossesType.ORPO:
            log_odds = (policy_chosen_logps - policy_rejected_logps) - (
                torch.log1p(-torch.clamp(torch.exp(policy_chosen_logps), max=1 - 1e-7))
                - torch.log1p(-torch.clamp(torch.exp(policy_rejected_logps), max=1 - 1e-7))
            )
            ratio = -F.logsigmoid(log_odds)
            or_loss = self.dpo_loss_registry.beta * ratio
            nll_loss = -policy_chosen_logps

            metrics[f'{prefix}orpo/nll_loss'] = nll_loss.detach().cpu().mean().item()
            metrics[f'{prefix}orpo/or_loss'] = or_loss.detach().cpu().mean().item()
            metrics[f'{prefix}orpo/ratio'] = (ratio).detach().cpu().mean().item()
            metrics[f'{prefix}orpo/log_odds'] = (log_odds).detach().cpu().mean().item()

        elif self.loss_type == DPOLossesType.ASFT:
            chosen_ratio = policy_chosen_logps - (
                torch.log1p(-torch.clamp(torch.exp(policy_chosen_logps), max=1 - 1e-7))
            )
            rejected_ratio = policy_rejected_logps - (
                torch.log1p(-torch.clamp(torch.exp(policy_rejected_logps), max=1 - 1e-7))
            )
            chosen_logsig = -F.logsigmoid(chosen_ratio)
            rejected_logsig = -F.logsigmoid(-rejected_ratio)

            asft_loss = self.dpo_loss_registry.beta * (chosen_logsig + rejected_logsig)
            nll_loss = -policy_chosen_logps

            metrics[f'{prefix}asft/nll_loss'] = nll_loss.detach().cpu().mean().item()
            metrics[f'{prefix}asft/asft_loss'] = asft_loss.detach().cpu().mean().item()
            metrics[f'{prefix}asft/chosen_logsig'] = chosen_logsig.detach().cpu().mean().item()
            metrics[f'{prefix}asft/rejected_logsig'] = rejected_logsig.detach().cpu().mean().item()

        if self.sft_model is not None:
            sft_chosen_logps, sft_rejected_logps = self._get_logps(self.sft_model, batch)

            with torch.no_grad():
                _, sft_chosen_rewards, sft_rejected_rewards = self.dpo_loss(
                    policy_chosen_logps=policy_chosen_logps,
                    policy_rejected_logps=policy_rejected_logps,
                    reference_chosen_logps=sft_chosen_logps,
                    reference_rejected_logps=sft_rejected_logps,
                    precomputed_margins=precomputed_margins,
                )

            sft_prefix_name = prefix + 'rewards/sft_'
            metrics = self._compute_metrics(metrics, sft_prefix_name, sft_chosen_rewards, sft_rejected_rewards)

        return losses.mean(), metrics

    def _compute_metrics(
        self, metrics: dict[str, float], prefix_name: str, chosen_rewards: torch.Tensor, rejected_rewards: torch.Tensor
    ) -> dict[str, float]:
        accuracies = (chosen_rewards > rejected_rewards).float()
        metrics[f'{prefix_name}chosen'] = (chosen_rewards).detach().cpu().mean().item()
        metrics[f'{prefix_name}rejected'] = (rejected_rewards).detach().cpu().mean().item()
        metrics[f'{prefix_name}margins'] = (chosen_rewards - rejected_rewards).detach().cpu().mean().item()
        metrics[f'{prefix_name}accuracies'] = accuracies.detach().cpu().mean().item()

        metrics[f'{prefix_name}grad_term'] = (
            (self.dpo_loss_registry.beta * F.sigmoid(rejected_rewards - chosen_rewards)).detach().cpu().mean().item()
        )
        metrics[f'{prefix_name}grad_term_std'] = (
            (self.dpo_loss_registry.beta * F.sigmoid(rejected_rewards - chosen_rewards)).detach().cpu().std().item()
        )

        return metrics

    def _compute_flips(
        self,
        metrics: dict[str, Any],
        prefix_name: str,
        logp_accuracies: torch.Tensor,
        ref_logp_accuracies: torch.Tensor,
    ):
        correct_correct = (ref_logp_accuracies == 1) & (logp_accuracies == 1)
        correct_incorrect = (ref_logp_accuracies == 1) & (logp_accuracies == 0)
        incorrect_correct = (ref_logp_accuracies == 0) & (logp_accuracies == 1)
        incorrect_incorrect = (ref_logp_accuracies == 0) & (logp_accuracies == 0)

        correct_correct_count = correct_correct.sum().item()
        correct_incorrect_count = correct_incorrect.sum().item()
        incorrect_correct_count = incorrect_correct.sum().item()
        incorrect_incorrect_count = incorrect_incorrect.sum().item()

        total_count = len(logp_accuracies)

        correct_correct_ratio = correct_correct_count / total_count
        correct_incorrect_ratio = correct_incorrect_count / total_count
        incorrect_correct_ratio = incorrect_correct_count / total_count
        incorrect_incorrect_ratio = incorrect_incorrect_count / total_count

        metrics[f'{prefix_name}flips/correct->correct'] = correct_correct_ratio
        metrics[f'{prefix_name}flips/correct->incorrect'] = correct_incorrect_ratio
        metrics[f'{prefix_name}flips/incorrect->correct'] = incorrect_correct_ratio
        metrics[f'{prefix_name}flips/incorrect->incorrect'] = incorrect_incorrect_ratio

        return metrics

    def compute_loss(
        self,
        model: PreTrainedModel | nn.Module,
        inputs: dict[str, torch.Tensor | Any],
        return_outputs=False,
    ) -> torch.Tensor | tuple[torch.Tensor, dict[str, float]]:
        loss, metrics = self.get_batch_metrics(model, inputs, train_eval='train')

        if self.accelerator.is_main_process:
            self.store_metrics(metrics, train_eval='train')
        if return_outputs:
            return loss, metrics
        return loss

    def prediction_step(
        self,
        model: PreTrainedModel | nn.Module,
        inputs: dict[str, torch.Tensor | Any],
        prediction_loss_only: bool,
        ignore_keys: list[str] | None = None,
    ) -> tuple[torch.Tensor, torch.Tensor | None, torch.Tensor | None]:
        if ignore_keys is None:
            if hasattr(model, 'config'):
                ignore_keys = getattr(model.config, 'keys_to_ignore_at_inference', [])
            else:
                ignore_keys = []

        with torch.no_grad():
            loss, metrics = self.get_batch_metrics(model, inputs, train_eval='eval')

        if self.accelerator.is_main_process:
            self.store_metrics(metrics, train_eval='eval')
        if prediction_loss_only:
            return loss.detach(), None, None

        logits_dict = {
            'logits_test/chosen': metrics['logits_test/chosen'],
            'logits_test/rejected': metrics['logits_test/rejected'],
        }
        logits = tuple(v for k, v in logits_dict.items() if k not in ignore_keys)
        logits = torch.stack(logits).mean(axis=1)  # type: ignore[call-overload, arg-type]
        labels = torch.zeros(logits.shape[0])

        return loss.detach(), logits, labels

    def store_metrics(self, metrics: dict[str, float], train_eval: Literal['train', 'eval'] = 'train') -> None:
        for key, value in metrics.items():
            self._stored_metrics[train_eval][key].append(value)

    def log(self, logs: dict[str, float]) -> None:
        train_eval = 'train' if 'loss' in logs else 'eval'
        for key, metrics in self._stored_metrics[train_eval].items():
            logs[key] = torch.tensor(metrics).cpu().mean().item()
        del self._stored_metrics[train_eval]
        return super().log(logs)  # pylint: disable=no-member



import torch

from turbo_alignment.modeling.imagebind import ImageBindSingleton
from turbo_alignment.modeling.imagebind.models import ImageBindSettings
from turbo_alignment.modeling.imagebind.models import (
    ModalityType as ImageBindModalityType,
)
from turbo_alignment.modeling.multimodal.encoders.image.base import BaseImageEncoder
from turbo_alignment.modeling.multimodal.encoders.registry import (
    ModalityEncoderRegistry,
)
from turbo_alignment.settings.modality import ModalityEncoderType


@ModalityEncoderRegistry.register(ModalityEncoderType.IMAGEBIND_IMAGE)
class ImageBindImageModeling(BaseImageEncoder):
    def __init__(self, imagebind_settings: ImageBindSettings | None = None):
        super().__init__()

        if imagebind_settings is None:
            imagebind_settings = ImageBindSettings()
        self.model_imagebind = ImageBindSingleton(imagebind_settings).get()

    def encode(self, inputs: torch.Tensor) -> torch.Tensor:
        input_values = {
            ImageBindModalityType.VISION: inputs,
        }
        image_embeddings = self.model_imagebind(input_values)[ImageBindModalityType.VISION]
        image_embeddings = torch.nn.functional.normalize(image_embeddings, dim=-1)
        return image_embeddings

    @property
    def emb_dim(self) -> int:
        return self.model_imagebind.modality_heads[ImageBindModalityType.VISION][-1].out_features



from typing import Any

from transformers import (
    TrainerCallback,
    TrainerControl,
    TrainerState,
    TrainingArguments,
)
from wandb.sdk.lib.disabled import RunDisabled
from wandb.sdk.wandb_run import Run

from turbo_alignment.common.logging import get_project_logger

logger = get_project_logger()


class BaseWandbCallback(TrainerCallback):
    def __init__(self, wandb_run: Run | RunDisabled) -> None:
        super().__init__()
        self._wandb_run = wandb_run

    def on_log(
        self,
        _args: TrainingArguments,
        state: TrainerState,
        _control: TrainerControl,
        **kwargs,
    ) -> None:
        logs = kwargs.get('logs', {})
        self._log(logs=logs, state=state)

    def _log(self, logs: dict[str, Any], state: TrainerState) -> None:
        rewritten_logs = self._rewrite_logs(logs)
        self._wandb_run.log({**rewritten_logs, 'train/global_step': state.global_step}, step=state.global_step)

    @staticmethod
    def _rewrite_logs(logs: dict[str, Any]) -> dict[str, Any]:
        rewritten_logs = {}
        eval_prefix = 'eval_'
        eval_prefix_len = len(eval_prefix)
        test_prefix = 'test_'
        test_prefix_len = len(test_prefix)
        train_prefix = 'train_'
        train_prefix_len = len(train_prefix)
        cherry_pick_prefix = 'cherry_pick_'
        cherry_pick_prefix_len = len(cherry_pick_prefix)
        for k, v in logs.items():
            if k.startswith(eval_prefix):
                rewritten_logs['eval/' + k[eval_prefix_len:]] = v
            elif k.startswith(test_prefix):
                rewritten_logs['test/' + k[test_prefix_len:]] = v
            elif k.startswith(train_prefix):
                rewritten_logs['train/' + k[train_prefix_len:]] = v
            elif k.startswith(cherry_pick_prefix):
                rewritten_logs['cherry_pick/' + k[cherry_pick_prefix_len:]] = v
            else:
                rewritten_logs['train/' + k] = v

        return rewritten_logs



from abc import ABC
from pathlib import Path
from typing import Any, overload

import numpy as np
import torch
from safetensors._safetensors_rust import SafetensorError

from turbo_alignment.common.data.io import read_jsonl
from turbo_alignment.common.data.multimodal import BaseModalityReader
from turbo_alignment.common.data.multimodal.registry import ModalityReaderRegistry
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.common.registry import Params
from turbo_alignment.constants import DISABLE_LOSS_LABEL
from turbo_alignment.dataset.base import AlignmentDataset
from turbo_alignment.dataset.chat.chat import InferenceChatDataset, TrainChatDataset
from turbo_alignment.dataset.chat.models import ChatDatasetRecord, ChatMessage
from turbo_alignment.dataset.multimodal.models import (
    MultimodalDatasetRecord,
    MultimodalFileMessage,
    MultimodalTextMessage,
)
from turbo_alignment.dataset.registry import MultimodalDatasetTypeRegistry
from turbo_alignment.settings.datasets import DatasetStrategy
from turbo_alignment.settings.modality import Modality

VERBOSE_EVERY = 1000

logger = get_project_logger()


class MultimodalDataset(AlignmentDataset[MultimodalDatasetRecord], ABC):
    def __init__(self, tokenizer, source, settings):
        super().__init__(tokenizer=tokenizer, source=source, settings=settings)

        self._n_modality_embeddings = settings.n_modality_embeddings
        self._modality_reader_settings_mapping = settings.modality_reader_settings_mapping
        self._modality_token_mapping = settings.modality_token_mapping
        self._start_modality_token = settings.start_modality_token
        self._end_modality_token = settings.end_modality_token
        self._truncate_top = settings.truncate_top

        self._modality_readers: dict[Modality, BaseModalityReader] = {
            modality: ModalityReaderRegistry.by_name(modality).from_params(
                Params(
                    {
                        'type': self._modality_reader_settings_mapping[modality].reader_type,
                        'reader_path': self._modality_reader_settings_mapping[modality].reader_path,
                    }
                )
            )
            for modality in Modality
            if modality != Modality.TEXT and self._modality_reader_settings_mapping[modality]
        }

    def _tokenize(self, text: str) -> np.ndarray:
        return self.tokenizer(
            text,
            add_special_tokens=False,
            padding=False,
            truncation=False,
            return_tensors='np',
        )[
            'input_ids'
        ][0]

    def _get_token_id(self, token: str) -> int:
        encoded_token = self._tokenize(token)
        assert len(encoded_token) == 1
        return encoded_token[0]

    @staticmethod
    @overload
    def _read_records(records: Path) -> list[MultimodalDatasetRecord]:
        ...

    @staticmethod
    @overload
    def _read_records(records: list[dict]) -> list[MultimodalDatasetRecord]:
        ...

    @staticmethod
    def _read_records(records) -> list[MultimodalDatasetRecord]:
        if isinstance(records, Path):
            return [MultimodalDatasetRecord(**record) for record in read_jsonl(records)]
        if isinstance(records, list):
            return [MultimodalDatasetRecord(**record) for record in records]
        raise NotImplementedError

    def __get_modality_message(self, modality: Modality) -> str:
        assert modality != Modality.TEXT
        modality_token = self._modality_token_mapping[modality]
        modality_message_span = ''.join(modality_token for _ in range(self._n_modality_embeddings))
        return f'{self._start_modality_token}{modality_message_span}{self._end_modality_token}'

    def _convert_to_chat(self, record: MultimodalDatasetRecord) -> ChatDatasetRecord:
        """
        Обычные текстовые сообщения оставляем без изменений, а мультимодальные сообщения
        преобразуем в <M><modality_tag><modality_tag>...</M>.
        <modality_tag> повторяется n_modality_embeddings раз.
        """

        converted_messages: list[ChatMessage] = []
        for msg in record.messages:
            if isinstance(msg, MultimodalTextMessage):
                converted_messages.append(
                    ChatMessage(role=msg.role, content=msg.content, disable_loss=msg.disable_loss)
                )
            else:
                converted_messages.append(
                    ChatMessage(
                        role=msg.role,
                        content=self.__get_modality_message(msg.type),
                        disable_loss=msg.disable_loss,
                    )
                )

        return ChatDatasetRecord(id=record.id, messages=converted_messages)

    def _read_modalities(
        self, record: MultimodalDatasetRecord, modality_messages_after_truncation: int
    ) -> list[tuple[Modality, torch.Tensor]]:
        modality_messages: list[MultimodalFileMessage] = [
            m for m in record.messages if isinstance(m, MultimodalFileMessage)
        ]

        messages_to_delete = len(modality_messages) - modality_messages_after_truncation

        if self._truncate_top:
            modality_messages = modality_messages[messages_to_delete:]
        else:
            modality_messages = modality_messages[:modality_messages_after_truncation]

        modality_encodings: list[tuple[Modality, torch.Tensor]] = []
        for msg in modality_messages:
            reader = self._modality_readers[msg.type]
            modality_encodings.append((msg.type, reader.read(msg.content)))
        return modality_encodings


@MultimodalDatasetTypeRegistry.register(DatasetStrategy.TRAIN)
class TrainMultimodalDataset(MultimodalDataset):
    def __init__(self, tokenizer, source, settings) -> None:
        """

        :param n_modality_embeddings: сколько токенов выделяем под одно сообщение с нетекстовой модальностью
        :param modality_token_mapping: modality -> token
        :param start_modality_token: начало блока с токенами, которые относятся к нетекстовой модальности
        :param end_modality_token: конец блока с токенами, которые относятся к нетекстовой модальности
        :param truncate_top: обрезаем диалог сверху или снизу, если он не влазит в max_tokens
        """

        super().__init__(tokenizer=tokenizer, source=source, settings=settings)

        self._chat_dataset = TrainChatDataset(tokenizer=tokenizer, source=source, settings=settings, read=False)

        self._read()

    def convert_records(self, records: list[MultimodalDatasetRecord]) -> list[dict[str, Any] | None]:
        chat_records = [self._convert_to_chat(r) for r in records]
        tokenized_chat_records = self._chat_dataset.convert_records(chat_records)

        outputs: list[dict[str, Any] | None] = []

        for i, (record, tokenized_record) in enumerate(zip(records, tokenized_chat_records)):
            if i % VERBOSE_EVERY == 0:
                logger.info(f'{i} records processed of {self.source.name}')

            if tokenized_record is None:
                outputs.append(None)
                continue

            modality_messages_after_truncation = int(
                (tokenized_record['input_ids'] == self._get_token_id(self._start_modality_token)).sum()
            )

            try:
                encoded_modalities = self._read_modalities(record, modality_messages_after_truncation)
            except (OSError, RuntimeError, SafetensorError):
                outputs.append(None)
                continue

            if len(encoded_modalities) != modality_messages_after_truncation:
                outputs.append(None)
                continue

            modality_tokens_mask = torch.isin(
                tokenized_record['input_ids'],
                torch.tensor([self._get_token_id(token) for token in self._modality_token_mapping.values()]),
            )

            tokenized_record['labels'][modality_tokens_mask] = DISABLE_LOSS_LABEL

            outputs.append(
                {
                    **tokenized_record,
                    'modality_inputs': encoded_modalities,
                    'modality_tokens_mask': modality_tokens_mask,
                }
            )

        return outputs


@MultimodalDatasetTypeRegistry.register(DatasetStrategy.INFERENCE)
class InferenceMultimodalDataset(MultimodalDataset):
    def __init__(
        self,
        *args,
        random_cut: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        settings = kwargs['settings']
        settings.random_cut = random_cut
        self._chat_dataset = InferenceChatDataset(
            tokenizer=kwargs['tokenizer'], source=kwargs['source'], settings=settings, read=False
        )
        self._read()

    def convert_records(self, records: list[MultimodalDatasetRecord]) -> list[dict[str, Any] | None]:
        chat_records = [self._convert_to_chat(r) for r in records]
        tokenized_chat_records = self._chat_dataset.convert_records(chat_records)

        outputs: list[dict[str, Any] | None] = []

        for record, tokenized_record in zip(records, tokenized_chat_records):
            if tokenized_record is None:
                outputs.append(None)
                continue

            modality_messages_after_truncation = int(
                (tokenized_record['input_ids'] == self._get_token_id(self._start_modality_token)).sum()
            )

            try:
                encoded_modalities = self._read_modalities(record, modality_messages_after_truncation)
            except (OSError, RuntimeError, SafetensorError):
                outputs.append(None)
                continue

            if len(encoded_modalities) != modality_messages_after_truncation:
                outputs.append(None)
                continue

            modality_tokens_mask = torch.isin(
                tokenized_record['input_ids'],
                torch.tensor([self._get_token_id(token) for token in self._modality_token_mapping.values()]),
            )

            if 'labels' in tokenized_record:
                tokenized_record['labels'][modality_tokens_mask] = DISABLE_LOSS_LABEL

            modality_object_paths = [str(r.content) for r in record.messages if r.type in ('image', 'audio')]

            outputs.append(
                {
                    **tokenized_record,
                    'modality_inputs': encoded_modalities,
                    'modality_tokens_mask': modality_tokens_mask,
                    'modality_object_paths': modality_object_paths,
                }
            )
        return outputs



# mypy: ignore-errors
import copy
import inspect
from typing import Any, Callable, Generic, Type, TypeVar

from turbo_alignment.common.registry.params import Params

T = TypeVar('T')


class Lazy(Generic[T]):
    def __init__(
        self,
        constructor: Type[T] | Callable[..., T],
        params: Params | None = None,
        constructor_extras: dict[str, Any] | None = None,
        **kwargs,
    ) -> None:
        self._constructor = constructor
        self._params = params or Params({})
        self._constructor_extras = constructor_extras or {}
        self._constructor_extras.update(kwargs)

    @property
    def constructor(self) -> Callable[..., T]:
        if inspect.isclass(self._constructor):

            def constructor_to_use(**kwargs):
                return self._constructor.from_params(  # type: ignore[union-attr]
                    copy.deepcopy(self._params),
                    **kwargs,
                )

            return constructor_to_use

        return self._constructor

    def construct(self, **kwargs) -> T:
        contructor_kwargs = {**self._constructor_extras, **kwargs}
        return self.constructor(**contructor_kwargs)



from pathlib import Path
from typing import Any, overload

from transformers import PreTrainedTokenizerBase

from turbo_alignment.common.data.io import read_jsonl
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.dataset.base import AlignmentDataset
from turbo_alignment.dataset.chat import (
    ChatDatasetRecord,
    ChatMessage,
    TrainChatDataset,
)
from turbo_alignment.dataset.pair_preferences.models import PairPreferenceRecord
from turbo_alignment.dataset.registry import PairPreferenceDatasetTypeRegistry
from turbo_alignment.settings.datasets.base import (
    DatasetSourceSettings,
    DatasetStrategy,
)
from turbo_alignment.settings.datasets.pair_preference import (
    PairPreferenceDatasetSettings,
)

logger = get_project_logger()


@PairPreferenceDatasetTypeRegistry.register(DatasetStrategy.TRAIN)
class PairPreferenceDataset(AlignmentDataset[PairPreferenceRecord]):
    def __init__(
        self,
        source: DatasetSourceSettings,
        settings: PairPreferenceDatasetSettings,
        tokenizer: PreTrainedTokenizerBase,
        read: bool = True,
    ):
        self._add_labels = settings.add_labels
        settings.chat_settings.keep_end = True
        self._chat_dataset = TrainChatDataset(
            source=source,
            settings=settings.chat_settings,
            tokenizer=tokenizer,
            read=False,
        )
        super().__init__(source=source, settings=settings, tokenizer=tokenizer)

        if read:
            self._read()

    def convert_records(self, records: list[PairPreferenceRecord]) -> list[dict[str, Any] | None]:
        chosen_chat_records: list[ChatDatasetRecord] = []
        rejected_chat_records: list[ChatDatasetRecord] = []

        for record in records:
            context = [
                ChatMessage(role=message.role, content=message.content, disable_loss=True)
                for message in record.context
            ]

            chosen = ChatMessage(role=record.answer_w.role, content=record.answer_w.content)
            rejected = ChatMessage(role=record.answer_l.role, content=record.answer_l.content)

            chosen_chat_records.append(ChatDatasetRecord(id=record.id, messages=context + [chosen]))
            rejected_chat_records.append(ChatDatasetRecord(id=record.id, messages=context + [rejected]))

        tokenized_chosen_records = self._chat_dataset.convert_records(chosen_chat_records)
        tokenized_rejected_records = self._chat_dataset.convert_records(rejected_chat_records)

        output: list[dict[str, Any] | None] = []
        for record, chosen_record, rejected_record in zip(
            records, tokenized_chosen_records, tokenized_rejected_records
        ):
            if not (chosen_record and rejected_record):
                continue

            ignore_keys = ['precomputed_margin']
            if not self._add_labels:
                ignore_keys.append('labels')

            chosen_tokens = {k: v.squeeze(0) for k, v in chosen_record.items() if k not in ignore_keys}
            rejected_tokens = {k: v.squeeze(0) for k, v in rejected_record.items() if k not in ignore_keys}

            output.append(
                {
                    'id': record.id,
                    'inputs_w': chosen_tokens,
                    'inputs_l': rejected_tokens,
                    'precomputed_margin': record.precomputed_margin,
                }
            )

        return output

    @staticmethod
    @overload
    def _read_records(records: Path) -> list[PairPreferenceRecord]:
        ...

    @staticmethod
    @overload
    def _read_records(records: list[dict]) -> list[PairPreferenceRecord]:
        ...

    @staticmethod
    def _read_records(records) -> list[PairPreferenceRecord]:
        if isinstance(records, Path):
            return [PairPreferenceRecord(**record) for record in read_jsonl(records)]
        if isinstance(records, list):
            return [PairPreferenceRecord(**record) for record in records]
        raise NotImplementedError



from transformers import AutoTokenizer
from transformers.tokenization_utils_base import PreTrainedTokenizerBase

from turbo_alignment.settings.model import PreTrainedModelSettings
from turbo_alignment.settings.tf.tokenizer import TokenizerSettings


def load_tokenizer(
    tokenizer_settings: TokenizerSettings,
    model_settings: PreTrainedModelSettings,
) -> PreTrainedTokenizerBase:
    tokenizer_path = tokenizer_settings.tokenizer_path or model_settings.model_path
    tokenizer = AutoTokenizer.from_pretrained(
        tokenizer_path,
        use_fast=tokenizer_settings.use_fast,
        trust_remote_code=tokenizer_settings.trust_remote_code,
        **tokenizer_settings.tokenizer_kwargs,
    )
    return tokenizer



from pathlib import Path

import typer

from turbo_alignment import pipelines
from turbo_alignment.cli.app import app
from turbo_alignment.settings.pipelines.inference.base import (
    InferenceExperimentSettings,
)
from turbo_alignment.settings.pipelines.inference.chat import (
    ChatInferenceExperimentSettings,
)
from turbo_alignment.settings.pipelines.inference.multimodal import (
    MultimodalInferenceExperimentSettings,
)
from turbo_alignment.settings.pipelines.inference.rag import (
    RAGInferenceExperimentSettings,
)


@app.command(name='inference_chat', help='Infer model on chat dataset')
def infer_chat_entrypoint(
    inference_settings_path: Path = typer.Option(
        ..., '--inference_settings_path', exists=True, help='Path to inference config file'
    )
) -> None:
    inference_settings = ChatInferenceExperimentSettings.parse_file(inference_settings_path)
    pipelines.ChatInferenceStrategy().run(inference_settings)


@app.command(name='inference_rm', help='Infer model on rm dataset')
def infer_rm_entrypoint(
    inference_settings_path: Path = typer.Option(
        ..., '--inference_settings_path', exists=True, help='Path to inference config file'
    )
) -> None:
    inference_settings = InferenceExperimentSettings.parse_file(inference_settings_path)
    pipelines.RMInferenceStrategy().run(inference_settings)


@app.command(name='inference_classification', help='Infer model on classification dataset')
def infer_classification_entrypoint(
    inference_settings_path: Path = typer.Option(
        ..., '--inference_settings_path', exists=True, help='Path to inference config file'
    )
) -> None:
    inference_settings = InferenceExperimentSettings.parse_file(inference_settings_path)
    pipelines.ClassificationInferenceStrategy().run(inference_settings)


@app.command(name='inference_multimodal', help='Infer model on chat dataset')
def infer_multimodal_entrypoint(
    inference_settings_path: Path = typer.Option(
        ..., '--inference_settings_path', exists=True, help='Path to inference config file'
    )
) -> None:
    inference_settings = MultimodalInferenceExperimentSettings.parse_file(inference_settings_path)
    pipelines.MultimodalInferenceStrategy().run(inference_settings)


@app.command(name='inference_rag', help='Infer model on classification dataset')
def infer_rag_entrypoint(
    inference_settings_path: Path = typer.Option(
        ..., '--inference_settings_path', exists=True, help='Path to inference config file'
    )
) -> None:
    inference_settings = RAGInferenceExperimentSettings.parse_file(inference_settings_path)
    pipelines.RAGInferenceStrategy().run(inference_settings)



from enum import Enum
from typing import Any

from pydantic import BaseModel, field_validator

from turbo_alignment.dataset.base.models import DatasetRecord


class ChatMessageRole(str, Enum):
    SYSTEM = 'system'
    USER = 'user'
    BOT = 'bot'


class ChatMessage(BaseModel):
    role: ChatMessageRole
    content: str
    disable_loss: bool = False

    @field_validator('role', mode='before')
    def set_bot_role(cls, values: str) -> str:
        if values == 'assistant':
            return ChatMessageRole.BOT
        return values


class ChatDatasetRecord(DatasetRecord):
    messages: list[ChatMessage]
    label: str | None = None
    meta: dict[str, Any] | None = None



import torch
from transformers import PreTrainedTokenizerBase

from turbo_alignment.dataset.multimodal.models import MultimodalDatasetRecord
from turbo_alignment.generators.base import ChatGeneratorBase
from turbo_alignment.settings.generators.chat import CustomChatGenerationSettings
from turbo_alignment.settings.generators.outputs.chat import AnswerMessage
from turbo_alignment.settings.generators.outputs.multimodal import (
    MultimodalInferenceOutput,
)
from turbo_alignment.settings.tf.generation import GeneratorTransformersSettings


class MultimodalGenerator(ChatGeneratorBase[MultimodalDatasetRecord, MultimodalInferenceOutput]):
    def __init__(
        self,
        transformers_settings: GeneratorTransformersSettings,
        custom_generation_settings: CustomChatGenerationSettings,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> None:
        super().__init__(
            transformers_settings=transformers_settings,
            custom_generation_settings=custom_generation_settings,
            tokenizer=tokenizer,
            **kwargs,
        )

    def _generate_from_batch_records(
        self,
        records: list[dict[str, torch.Tensor]],
        original_records: list[MultimodalDatasetRecord],
        dataset_name: str,
    ) -> list[MultimodalInferenceOutput]:
        raise ValueError('You can not use batch generation with multimodаl generator')

    def _generate_from_single_record(
        self,
        record: dict[str, torch.Tensor],
        original_record: MultimodalDatasetRecord,
        dataset_name: str,
    ) -> MultimodalInferenceOutput:
        input_ids = torch.unsqueeze(record['input_ids'], 0).to(self._model.language_model.device)
        attention_mask = torch.unsqueeze(record['attention_mask'], 0).to(self._model.language_model.device)
        modality_inputs = [record['modality_inputs']]
        modality_tokens_mask = record['modality_tokens_mask'].unsqueeze(0).to(self._model.language_model.device)

        inputs_embeds = self._model.convert_inputs_to_embeds(input_ids, modality_inputs, modality_tokens_mask)

        output_indices = self._model.language_model.generate(
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            tokenizer=self._tokenizer,
            generation_config=self._transformers_generator_parameters,
        )

        answers = self._decode(token_indices=output_indices)

        logits: torch.Tensor | None = None
        input_token_ids: torch.Tensor | None = None
        answer_tokens_ids: torch.Tensor | None = None

        if self._return_logits:
            with torch.no_grad():
                logits = self._model(output_indices).logits

            answer_tokens_ids = output_indices
            input_token_ids = input_ids

        return MultimodalInferenceOutput(
            id=original_record.id,
            dataset_name=dataset_name,
            messages=original_record.messages,
            label=original_record.label,
            answers=[
                AnswerMessage(
                    id=str(i),
                    content=a,
                    input_token_ids=input_token_ids,
                    answer_token_ids=answer_tokens_ids,
                    logits=logits,
                )
                for i, a in enumerate(answers)
            ],
        )



from typing import Literal

from turbo_alignment.settings.datasets.base import (
    BaseDatasetSettings,
    DatasetType,
    MultiDatasetSettings,
)
from turbo_alignment.settings.datasets.chat import ChatDatasetSettings
from turbo_alignment.settings.datasets.pair_preference import (
    PairPreferenceDatasetSettings,
)


class DDPODatasetSettings(BaseDatasetSettings):
    dataset_type: Literal[DatasetType.DDPO] = DatasetType.DDPO
    chat_settings: ChatDatasetSettings
    pair_preferences: PairPreferenceDatasetSettings


class DDPOMultiDatasetSettings(DDPODatasetSettings, MultiDatasetSettings):
    ...



from turbo_alignment.common.registry import Registrable
from turbo_alignment.settings.modality import Modality


class ModalityReaderRegistry(Registrable):
    ...


@ModalityReaderRegistry.register(Modality.AUDIO)
class AudioModalityReaderRegistry(Registrable):
    ...


@ModalityReaderRegistry.register(Modality.IMAGE)
class ImageModalityReaderRegistry(Registrable):
    ...
